#! /usr/env/python
"""

Landlab model of 2D diffusion, using DiffusionComponent.

Created July 2013 GT
Last updated August 2013 GT

"""

import sys                                    # for command-line arguments
from landlab import ModelParameterDictionary  # for input file
from landlab import create_and_initialize_grid
from landlab import RasterModelGrid
from landlab.plot.imshow import imshow_node_grid
import diffusion

class DiffusionModel(object):
    
    def initialize(self, input_stream=None):
    
        # If no input file/stream specified, use default (or command line?)
        # (or default values?)
        if input_stream is None:
            input_stream = str(raw_input('Enter name of input file: '))
        
        # Open input file
        inputs = ModelParameterDictionary(input_stream)
    
        # Create grid
        self.grid = create_and_initialize_grid(inputs)
        
        # Create a diffusion component
        self.diffusion_component = diffusion.DiffusionComponent(self.grid)
        self.diffusion_component.initialize(input_stream)
        
        # Read parameters
        self.run_duration = inputs.get('RUN_DURATION', ptype=float)
        self.opt_netcdf_output = inputs.get('OPT_FILE_OUTPUT', ptype='bool')
        self.opt_display_output = inputs.get('OPT_DISPLAY_OUTPUT', ptype='bool')
        
        self.setup_output_timing(inputs)
        
        
    def setup_output_timing(self, inputs):
        
        # Setup output timing
        if self.opt_netcdf_output:
            self.netcdf_output_interval = inputs.get('FILE_OUTPUT_INTERVAL',
                                                     ptype=float)
            self.next_file_output = self.netcdf_output_interval
        else:
            self.next_file_output = self.run_duration+1.0
            
        if self.opt_display_output:
            self.display_output_interval = inputs.get('DISPLAY_OUTPUT_INTERVAL',
                                                      ptype=float)
            self.next_display_output = self.display_output_interval
        else:
            self.next_display_output = self.run_duration+1.0
            
        self.find_next_stop_time()
                    
        # Time
        self.current_time = 0.0
        
        
    def handle_output(self):
        
        if self.current_time >= self.next_display_output:
            self.do_display_output()
            self.next_display_output += self.display_output_interval
            
        if self.current_time >= self.next_file_output:
            self.do_file_output()
            self.next_file_output += self.file_output_interval
            
            
    def find_next_stop_time(self):
        
        self.next_stop_time = min(self.run_duration, self.next_file_output,
                                  self.next_display_output)
            
            
    def do_display_output(self):
        
        # ok, here one problem is that we don't know what grid type we have
        if type(self.grid) is RasterModelGrid:
            print 'This is a raster grid'
        else:
            print 'non-raster grid'
            
        imshow_node_grid(self.grid, self.diffusion_component.z)
        
        
    def do_file_output(self):
        
        print 'File output goes here'
        
        
    def update(self):
        
        self.diffusion_component.run_one_step()
        
        
    def run(self):
        
        while self.current_time < self.run_duration:
            self.diffusion_component.run_until(self.next_stop_time)
            self.current_time = self.next_stop_time
            print self.current_time
            self.handle_output()
            self.find_next_stop_time()
        
        
    def finalize(self):
        
        pass
        
        
def main():
    
    # Instantiate the model
    difmod = DiffusionModel()

    # Handle command-line arguments (if any)
    if len(sys.argv)==1:
        input_file_name = None
    else:
        input_file_name = sys.argv[1]
    
    # Initialize the model
    difmod.initialize(input_file_name)

    # Run the model
    difmod.run()
    
    # Finalize
    difmod.finalize()
    


if __name__ == "__main__":
    main()
    
    
# Issues arising (for group discussion):
#  - how to get argument for input source to model: command-line argument?
#    command-line prompt? window??? is there a way to toggle between file and
#    command-line input? How best to be compatible with CMF/CMT?
#  - naming conventions for components, models, etc.
#  - How to design for maximum simplicity?
#  - How to get parameters/data passed between functions? Use class with data
#    members? Use big Data object?
#  - We need a way to handle defaults in MPD
#  - We need a way to create and return a grid whose subclass is determined by
#    the input file specs, not hardwired
#  - Better error handling for MPD
#  - Exception handling generally!

# Sketch of a class-based model:



import numpy as np
import pylab
from landlab import RasterModelGrid
from random import uniform
from landlab.components.simple_power_law_incision.power_law_fluvial_eroder import PowerLawIncision
from landlab.components.flow_routing.flow_routing_D8 import RouteFlowD8
from landlab.components.flow_accum.flow_accumulation2 import AccumFlow
import matplotlib.pyplot as plt

def main():
    nr = 50
    nc = 60
    nnodes = nr*nc
    dx=10
    #instantiate grid
    rg = RasterModelGrid(nr, nc, dx)
    rg.set_inactive_boundaries(False, True, True, True)
    
    z  = np.zeros( nnodes )    
    #set-up interior elevations with random numbers
    for i in range(0, nnodes):
        if rg.is_interior(i):
            z[i]=2+uniform(-0.5,0.5)
    
    #some helper and parameter values
    total_run_time = 10000 #yr
    one_twentieth_time = total_run_time/20
    uplift_rate = 0.001 #m/yr
    rain_rate = 1 #m/yr
    storm_duration = 50 #years
    
    elapsed_time = 0 #years
    
    #set up fluvial incision component
    incisor = PowerLawIncision('input_file.txt',rg)
    
    #print "elevations before ", rg.node_vector_to_raster(z)
    #interior_nodes are the nodes on which you will be operating
    interior_nodes = rg.get_active_cell_node_ids()
    
    while elapsed_time < total_run_time:  
        #erode the landscape
        z = incisor.run_one_storm(rg,z,rain_rate,storm_duration)
        #uplift the landscape
        z[interior_nodes] = z[interior_nodes]+uplift_rate * storm_duration
        #update the time
        elapsed_time = elapsed_time+storm_duration
        #print "elapsed_time", elapsed_time
        if elapsed_time%one_twentieth_time == 0:
            print "elapsed time",elapsed_time
            elev_raster = rg.node_vector_to_raster(z,True)
            # Plot topography
            pylab.figure(22)
            im = pylab.imshow(elev_raster, cmap=pylab.cm.RdBu, extent=[0, nc*rg.dx, 0, nr*rg.dx])
            cb = pylab.colorbar(im)
            cb.set_label('Elevation (m)', fontsize=12)
            pylab.title('Topography')
            pylab.show()
    
    #below purely for plotting reasons
    #instantiate variable of type RouteFlowD8 Class
    flow_router = RouteFlowD8(len(z))
    #initial flow direction
    flowdirs, max_slopes = flow_router.calc_flowdirs(rg,z)
    #insantiate variable of type AccumFlow Class
    accumulator = AccumFlow(rg)
    #initial flow accumulation
    drain_area = accumulator.calc_flowacc(rg, z, flowdirs)
    
    plt.loglog(np.array(drain_area),np.array(max_slopes),'ro',)
    plt.xlabel('drainage area, m')
    plt.ylabel('surface slope')
    plt.show()
    
    elev_raster = rg.node_vector_to_raster(z,True)
    # Plot topography
    pylab.figure(22)
    im = pylab.imshow(elev_raster, cmap=pylab.cm.RdBu, extent=[0, nc*rg.dx, 0, nr*rg.dx])
    cb = pylab.colorbar(im)
    cb.set_label('Elevation (m)', fontsize=12)
    pylab.title('Topography')
    
    pylab.show()    
    
    
if __name__ == '__main__':
    main()



#! /usr/bin/env python
import os
import warnings
import types


try:
    import netCDF4 as nc4
except ImportError:
    warnings.warn('Unable to import netCDF4.', ImportWarning)

from scipy.io import netcdf as nc


from landlab.io.netcdf._constants import (_AXIS_DIMENSION_NAMES,
                                          _AXIS_COORDINATE_NAMES,
                                          _NP_TO_NC_TYPE)

def _set_netcdf_attributes(root, attrs):
    """
    Set attributes of the netCDF Database object, *root*. Attributes are
    given as key/value pairs from *attrs*.
    """
    for (key, val) in attrs.items():
        setattr(root, key, val)


def _get_dimension_names(shape):
    names = _AXIS_DIMENSION_NAMES[- 1: - (len(shape) + 1): - 1]
    return names[::-1]


def _get_dimension_sizes(shape):
    names = _AXIS_DIMENSION_NAMES[- 1: - (len(shape) + 1): - 1]

    sizes = dict()
    for (axis, name) in enumerate(names):
        sizes[name] = shape[- (axis + 1)]

    return sizes


def _get_axes_names(shape):
    names = _AXIS_COORDINATE_NAMES[- 1: - (len(shape) + 1): - 1]
    return names[::-1]


def _set_netcdf_structured_dimensions(root, shape):
    """
    Add dimensions to *root* for a structured grid of size *shape*. The
    dimension names will be 'ni', 'nj', and 'nk'. 'ni' is the length of the
    fast dimension, followed by 'nj', and then 'nk'.

    For example, a grid with shape (3, 4, 5) will have dimensions ni=5,
    nj=4, and nk=3. Lower dimension grids simply drop the slowest dimension.
    Thus, a grid with shape (3, 4) has dimensions ni=4, and nj=3.
    """
    assert(len(shape) >= 1 and len(shape) <= 3)

    dimensions = _get_dimension_sizes(shape)

    dims = root.dimensions
    for (name, dim_size) in dimensions.items():
        if name not in dims:
            root.createDimension(name, dim_size)

    if not 'nt' in dims:
        nt = root.createDimension('nt', None)


def _set_netcdf_variables(root, fields, **kwds):
    """Set the field variables.
    
    First set the variables that define the grid and then the variables at
    the grid nodes and cells.
    """
    names = kwds.pop('names', None)

    _add_spatial_variables(root, fields, **kwds)
    #_add_variables_at_points(root, fields['nodes'])
    _add_variables_at_points(root, fields, names=names)


def _add_spatial_variables(root, grid, **kwds):
    """Add the variables to *root* that define the structured grid, *grid*.
    """
    long_name = kwds.get('long_name', {})

    vars = root.variables
    dims = root.dimensions

    spatial_variable_names = _get_axes_names(grid.shape)
    spatial_variable_shape = _get_dimension_names(grid.shape)

    for (axis, name) in enumerate(spatial_variable_names):
        try:
            var = vars[name]
        except KeyError:
            var = root.createVariable(name, 'f8', spatial_variable_shape)

        coords = grid.node_axis_coordinates(axis=axis).view()
        coords.shape = var.shape
        var[:] = coords

        var.units = grid.axis_units[axis]
        try:
            var.long_name = long_name[name]
        except KeyError:
            var.long_name = grid.axis_name[axis]


def _add_variables_at_points(root, fields, names=None):
    if isinstance(names, types.StringTypes):
        names = [names]
    names = names or fields['node'].keys()

    vars = root.variables

    spatial_variable_shape = _get_dimension_names(fields.shape)

    try:
        n_times = len(vars['t']) - 1
    except KeyError:
        n_times = 0

    node_fields = fields['node']
    for var_name in names:
        try:
            var = vars[var_name]
        except KeyError:
            var = root.createVariable(
                var_name, _NP_TO_NC_TYPE[str(node_fields[var_name].dtype)],
                ['nt'] + spatial_variable_shape)

        if node_fields[var_name].size > 1:
            data = node_fields[var_name].view()
            data.shape = var.shape[1:]
            try:
                var[n_times, :] = data
            except ValueError:
                raise
        else:
            var[n_times] = node_fields[var_name].flat[0]

        var.units = node_fields.units[var_name] or '?'
        var.long_name = var_name


def _add_time_variable(root, time, **kwds):
    units = kwds.get('units', 'days')
    reference = kwds.get('reference', '00:00:00 UTC')

    vars = root.variables

    try:
        t = vars['t']
    except KeyError:
        t = root.createVariable('t', 'f8', ('nt', ))
        t.units = ' '.join([units, 'since', reference])
        t.long_name = 'time'

    n_times = len(t)
    if time is not None:
        t[n_times] = time
    else:
        t[n_times] = n_times


_VALID_NETCDF_FORMATS = set([
    'NETCDF3_CLASSIC',
    'NETCDF3_64BIT',
    'NETCDF4_CLASSIC',
    'NETCDF4',
])

def write_netcdf(path, fields, attrs=None, append=False,
                 format='NETCDF3_64BIT', names=None):
    """Write landlab fields to netcdf.

    Write the data and grid information for *fields* to *path* as NetCDF.
    If the *append* keyword argument in True, append the data to an existing
    file, if it exists. Otherwise, clobber an existing files.

    Parameters
    ----------
    path : str
        Path to output file.
    fields : field-like
        Landlab field object that holds a grid and associated values.
    append : boolean, optional
        Append data to an existing file, otherwise clobber the file.
    format : {'NETCDF3_CLASSIC', 'NETCDF3_64BIT', 'NETCDF4_CLASSIC', 'NETCDF4'}
        Format of output netcdf file.
    attrs : dict
        Attributes to add to netcdf file.
    names : iterable of str, optional
        Names of the fields to include in the netcdf file. If not provided,
        write all fields.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.netcdf import write_netcdf

    Create a uniform rectilinear grid with four rows and 3 columns, and add
    some data fields to it.

    >>> rmg = RasterModelGrid(4, 3)
    >>> _ = rmg.add_field('node', 'topographic_elevation', np.arange(12.))
    >>> _ = rmg.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    Create a temporary directory to write the netcdf file into.

    >>> import tempfile, os
    >>> temp_dir = tempfile.mkdtemp()
    >>> os.chdir(temp_dir)

    Write the grid to a netcdf3 file but only include the *uplift_rate*
    data in the file.

    >>> write_netcdf('test.nc', rmg, format='NETCDF3_64BIT',
    ...     names='uplift_rate')

    Read the file back in and check its contents.

    >>> from scipy.io import netcdf
    >>> fp = netcdf.netcdf_file('test.nc', 'r')
    >>> 'uplift_rate' in fp.variables
    True
    >>> 'topographic_elevation' in fp.variables
    False
    >>> fp.variables['uplift_rate'][:].flatten()
    array([  0.,   2.,   4.,   6.,   8.,  10.,  12.,  14.,  16.,  18.,  20.,
           22.])
    """
    if format not in _VALID_NETCDF_FORMATS:
        raise ValueError('format not understood')

    attrs = attrs or {}

    if os.path.isfile(path) and append:
        mode = 'a'
    else:
        mode = 'w'

    if format == 'NETCDF3_CLASSIC':
        root = nc.netcdf_file(path, mode, version=1)
    elif format == 'NETCDF3_64BIT':
        root = nc.netcdf_file(path, mode, version=2)
    else:
        root = nc4.Dataset(path, mode, format=format)

    _set_netcdf_attributes(root, attrs)
    _set_netcdf_structured_dimensions(root, fields.shape)
    _set_netcdf_variables(root, fields, names=names)

    root.close()



#! /usr/bin/env python

import numpy

from landlab.grid.base import ModelGrid, CORE_NODE, BAD_INDEX_VALUE
from scipy.spatial import Voronoi

def simple_poly_area(x, y):
    """Calculates and returns the area of a 2-D simple polygon.
    
    Input vertices must be in sequence (clockwise or counterclockwise). *x*
    and *y* are arrays that give the x- and y-axis coordinates of the
    polygon's vertices.

    Parameters
    ----------
    x : ndarray
        x-coordinates of of polygon vertices.
    y : ndarray
        y-coordinates of of polygon vertices.

    Returns
    -------
    out : float
        Area of the polygon

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.voronoi import simple_poly_area
    >>> x = np.array([3., 1., 1., 3.])
    >>> y = np.array([1.5, 1.5, 0.5, 0.5])
    >>> simple_poly_area(x, y)
    2.0

    If the input coordinate arrays are 2D, calculate the area of each polygon.
    Note that when used in this mode, all polygons must have the same
    number of vertices, and polygon vertices are listed column-by-column.

    >>> x = np.array([[ 3.,  1.,  1.,  3.],
    ...               [-2., -2., -1., -1.]]).T
    >>> y = np.array([[1.5, 1.5, 0.5, 0.5],
    ...               [ 0.,  1.,  2.,  0.]]).T
    >>> simple_poly_area(x, y)
    array([ 2. ,  1.5])
    """
    # For short arrays (less than about 100 elements) it seems that the
    # Python sum is faster than the numpy sum. Likewise for the Python
    # built-in abs.
    return .5 * abs(sum(x[:-1] * y[1:] - x[1:] * y[:-1]) +
                    x[-1] * y[0] - x[0] * y[-1])


def calculate_link_lengths(pts, link_from, link_to):
    """Calculates and returns length of links between nodes.

    Parameters
    ----------
    pts : Nx2 numpy array containing (x,y) values
    link_from : 1D numpy array containing index numbers of nodes at starting
                point ("from") of links
    link_to : 1D numpy array containing index numbers of nodes at ending point
              ("to") of links
                      
    Returns
    -------
    out : ndarray
        1D numpy array containing horizontal length of each link
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.voronoi import calculate_link_lengths
    >>> pts = np.array([[0.,0.],[3.,0.],[3.,4.]]) # 3:4:5 triangle
    >>> lfrom = np.array([0,1,2])
    >>> lto = np.array([1,2,0])
    >>> calculate_link_lengths(pts, lfrom, lto)
    array([ 3.,  4.,  5.])
    """
    dx = pts[link_to, 0] - pts[link_from, 0]
    dy = pts[link_to, 1] - pts[link_from, 1]
    link_length = numpy.sqrt(dx * dx + dy * dy)
    return link_length
    
    
class VoronoiDelaunayGrid(ModelGrid):
    """
    This inherited class implements an unstructured grid in which cells are
    Voronoi polygons and nodes are connected by a Delaunay triangulation. Uses
    scipy.spatial module to build the triangulation.
    
    Examples
    --------
    >>> from numpy.random import rand
    >>> from landlab.grid import VoronoiDelaunayGrid
    >>> x, y = rand(25), rand(25)
    >>> vmg = VoronoiDelaunayGrid(x, y)  # node_x_coords, node_y_coords
    >>> vmg.number_of_nodes
    25
    """
    def __init__(self, x=None, y=None, reorient_links=True, **kwds):
        """Create a Voronoi Delaunay grid from a set of points.

        Create an unstructured grid from points whose coordinates are given
        by the arrays *x*, *y*.

        Parameters
        ----------
        x : array_like
            x-coordinate of points
        y : array_like
            y-coordinate of points

        Returns
        -------
        VoronoiDelaunayGrid
            A newly-created grid.

        Examples
        --------
        >>> from numpy.random import rand
        >>> from landlab.grid import VoronoiDelaunayGrid
        >>> x, y = rand(25), rand(25)
        >>> vmg = VoronoiDelaunayGrid(x, y)  # node_x_coords, node_y_coords
        >>> vmg.number_of_nodes
        25
        """
        if (x is not None) and (y is not None):
            self._initialize(x, y, reorient_links)
        super(VoronoiDelaunayGrid, self).__init__(**kwds)
        
    def _initialize(self, x, y, reorient_links=False):
        """
        Creates an unstructured grid around the given (x,y) points.
        """
        
        assert type(x)==numpy.ndarray, 'x must be a numpy array'
        assert type(y)==numpy.ndarray, 'y must be a numpy array'
        assert len(x)==len(y), 'x and y arrays must have the same size'
        
        # Make a copy of the points in a 2D array (useful for calls to geometry
        # routines, but takes extra memory space).
        pts = numpy.zeros((len(x), 2))
        pts[:,0] = x
        pts[:,1] = y
        self.pts = pts
        
        # NODES AND CELLS: Set up information pertaining to nodes and cells:
        #   - number of nodes
        #   - node x, y coordinates
        #   - default boundary status 
        #   - interior and boundary nodes
        #   - nodes associated with each cell and active cell
        #   - cells and active cells associated with each node 
        #     (or BAD_VALUE_INDEX if none)
        #
        # Assumptions we make here:
        #   - all interior (non-perimeter) nodes have cells (this should be 
        #       guaranteed in a Delaunay triangulation, but there may be 
        #       special cases)
        #   - all cells are active (later we'll build a mechanism for the user
        #       specify a subset of cells as active)
        #
        self._num_nodes = len(x)
        #print x, y
        self._node_x = x
        self._node_y = y
        [self.node_status, self._core_nodes, self._boundary_nodes] = \
                self.find_perimeter_nodes(pts)
        self._num_active_nodes = self.number_of_nodes
        self._num_core_nodes = len(self.core_nodes)
        self._num_cells = len(self.core_nodes)
        self._num_active_cells = self.number_of_cells
        [self.node_cell, self.cell_node] = self.setup_node_cell_connectivity(
            self.node_status, self.number_of_cells)
        self.node_activecell = self.node_cell
        self.activecell_node = self.cell_node

        # ACTIVE CELLS: Construct Voronoi diagram and calculate surface area of
        # each active cell.
        vor = Voronoi(pts)
        self.vor = vor
        self.active_cell_areas = numpy.zeros(self.number_of_active_cells)
        for node in self.activecell_node:
            xv = vor.vertices[vor.regions[vor.point_region[node]],0]
            yv = vor.vertices[vor.regions[vor.point_region[node]],1]
            self.active_cell_areas[self.node_activecell[node]] = simple_poly_area(xv, yv)
        
        # LINKS: Construct Delaunay triangulation and construct lists of link
        # "from" and "to" nodes.
        (self.link_fromnode,
         self.link_tonode,
         self.active_links_ids,
         self.face_width) = self.create_links_and_faces_from_voronoi_diagram(vor)
        
        # Optionally re-orient links so that they all point within upper-right
        # semicircle
        if reorient_links:
            self.reorient_links_upper_right()

        #[self.link_fromnode, self.link_tonode, self.active_links, self.face_width] \
        #        = self.create_links_and_faces_from_voronoi_diagram(vor)
        self._num_links = len(self.link_fromnode)
        self._num_faces = self._num_links # temporary: to be done right!
                    
        # LINKS: Calculate link lengths
        self._link_length = calculate_link_lengths(pts, self.link_fromnode,
                                                  self.link_tonode)
                                                       
        # LINKS: inlink and outlink matrices
        self._setup_inlink_and_outlink_matrices()
        
        # ACTIVE LINKS: Create list of active links, as well as "from" and "to"
        # nodes of active links.
        self._reset_list_of_active_links()

        # LINKS: set up link unit vectors and node unit-vector sums
        self._make_link_unit_vectors()

        # LINKS: ID of corresponding face, if any
        self.link_face = (numpy.zeros(self.number_of_links, dtype=int) +
                          BAD_INDEX_VALUE)  # make the list
        face_id = 0
        for link in self.active_links:
            self.link_face[link] = face_id
            face_id += 1
            
    @property
    def number_of_patches(self):
        """Number of patches.
        Returns the number of patches over the grid.
        """
        try:
            return self._number_of_patches
        except AttributeError:
            self.create_patches_from_delaunay_diagram(self.pts, self.vor)
            return self._number_of_patches
            
    @property
    def patch_nodes(self):
        """patch_nodes
        Returns the four nodes at the corners of each patch in a regular grid.
        """
        try:
            return self._patch_nodes
        except AttributeError:
            self.create_patches_from_delaunay_diagram(self.pts, self.vor)
            return self._patch_nodes

    def node_patches(self, nodata=-1):
        """node_patches()
        (This is a placeholder method until improved using jagged array 
        operations.)
        Returns a (N,max_voronoi_polygon_sides) array of the patches associated 
        with each node in the grid.
        The patches are returned in id order, with any null or nonexistent 
        patches recorded after the ids of existing faces.
        The nodata argument allows control of the array value used to indicate
        nodata. It defaults to -1, but other options are 'nan' and 'bad_value'.
        Note that this method returns a *masked* array, with the normal provisos
        that integer indexing with a masked array removes the mask.
        """
        if nodata==-1: #fiddle needed to ensure we set the nodata value properly if we've called patches elsewhere
            try:
                return self._node_patches
            except AttributeError:
                self.create_patches_from_delaunay_diagram(self.pts, self.vor, nodata)        
                return self._node_patches
        else:
            try:
                self.set_bad_value
            except:
                self.create_patches_from_delaunay_diagram(self.pts, self.vor, nodata)  
                self.set_bad_value=True      
                return self._node_patches
            else:
                return self._node_patches
            

    def find_perimeter_nodes(self, pts):
        """
        Uses a convex hull to locate the perimeter nodes of the Voronoi grid,
        then sets them as fixed value boundary nodes.
        It then sets/updates the various relevant node lists held by the grid, 
        and returns *node_status*, *core_nodes*, *boundary_nodes*.
        """
    
        # Calculate the convex hull for the set of points
        from scipy.spatial import ConvexHull
        hull = ConvexHull(pts, qhull_options='Qc') # see below why we use 'Qt'
        
        # The ConvexHull object lists the edges that form the hull. We need to
        # get from this list of edges the unique set of nodes. To do this, we
        # first flatten the list of vertices that make up all the hull edges 
        # ("simplices"), so it becomes a 1D array. With that, we can use the set()
        # function to turn the array into a set, which removes duplicate vertices.
        # Then we turn it back into an array, which now contains the set of IDs for
        # the nodes that make up the convex hull.
        #   The next thing to worry about is the fact that the mesh perimeter 
        # might contain nodes that are co-planar (that is, co-linear in our 2D 
        # world). For example, if you make a set of staggered points for a
        # hexagonal lattice using make_hex_points(), there will be some 
        # co-linear points along the perimeter. The ones of these that don't 
        # form convex corners won't be included in convex_hull_nodes, but they
        # are nonetheless part of the perimeter and need to be included in
        # the list of boundary_nodes. To deal with this, we pass the 'Qt'
        # option to ConvexHull, which makes it generate a list of coplanar
        # points. We include these in our set of boundary nodes.
        convex_hull_nodes = numpy.array(list(set(hull.simplices.flatten())))
        coplanar_nodes = hull.coplanar[:,0]
        boundary_nodes = numpy.concatenate((convex_hull_nodes, coplanar_nodes))
    
        # Now we'll create the "node_status" array, which contains the code
        # indicating whether the node is interior and active (=0) or a
        # boundary (=1). This means that all perimeter (convex hull) nodes are
        # initially flagged as boundary code 1. An application might wish to change
        # this so that, for example, some boundaries are inactive.
        node_status = numpy.zeros(len(pts[:,0]), dtype=numpy.int8)
        node_status[boundary_nodes] = 1
        
        # It's also useful to have a list of interior nodes
        core_nodes = numpy.where(node_status==0)[0]
        
        #save the arrays and update the properties
        self.node_status = node_status
        self._num_active_nodes = node_status.size
        self._num_core_nodes = len(core_nodes)
        self._num_core_cells = len(core_nodes)
        self.core_cells = numpy.arange(len(core_nodes))
        self.node_corecell = numpy.empty(node_status.size)
        self.node_corecell.fill(BAD_INDEX_VALUE)
        self.node_corecell[core_nodes] = self.core_cells
        self.active_cells = numpy.arange(node_status.size)
        self.cell_node = core_nodes
        self.activecell_node = core_nodes
        self.corecell_node = core_nodes
        self._boundary_nodes = boundary_nodes
        
        # Return the results
        return node_status, core_nodes, boundary_nodes

    @staticmethod
    def setup_node_cell_connectivity(node_status, ncells):
        """Setup node connectivity

        Creates and returns the following arrays:
        1. For each node, the ID of the corresponding cell, or
           BAD_INDEX_VALUE if the node has no cell.
        2. For each cell, the ID of the corresponding node.
            
        Parameters
        ----------
        node_status : ndarray of ints
            1D array containing the boundary status code for each node.
        ncells : ndarray of ints
            Number of cells (must equal the number of occurrences of CORE_NODE
            in node_status).
                    
        Examples
        --------
        >>> import numpy as np
        >>> from landlab.grid import VoronoiDelaunayGrid, BAD_INDEX_VALUE
        >>> ns = np.array([1, 0, 0, 1, 0])  # 3 interior, 2 boundary nodes
        >>> [node_cell, cell_node] = VoronoiDelaunayGrid.setup_node_cell_connectivity(ns, 3)
        >>> node_cell[1:3]
        array([0, 1])
        >>> node_cell[0] == BAD_INDEX_VALUE
        True
        >>> cell_node
        array([1, 2, 4])
        """
        assert ncells==numpy.count_nonzero(node_status==CORE_NODE), \
               'ncells must equal number of CORE_NODE values in node_status'

        cell = 0
        node_cell = numpy.ones(len(node_status), dtype=int)*BAD_INDEX_VALUE
        cell_node = numpy.zeros(ncells, dtype=int)
        for node in range(len(node_cell)):
            if node_status[node] == CORE_NODE:
                node_cell[node] = cell
                cell_node[cell] = node
                cell += 1
                
        #save the arrays
        #self.node_cell = node_cell
        #self.cell_node = cell_node
        
        return node_cell, cell_node
        
    @staticmethod
    def create_links_from_triangulation(tri):
        """
        From a Delaunay Triangulation of a set of points, contained in a
        scipy.spatial.Delaunay object "tri", creates and returns:
            1) a numpy array containing the ID of the "from" node for each link
            2) a numpy array containing the ID of the "to" node for each link
            3) the number of links in the triangulation
        
        Examples
        --------
        >>> from scipy.spatial import Delaunay
        >>> import numpy as np
        >>> from landlab.grid import VoronoiDelaunayGrid
        >>> pts = np.array([[ 0., 0.], [ 1., 0.], [ 1., 0.87],
        ...                 [-0.5, 0.87], [ 0.5, 0.87], [ 0., 1.73],
        ...                 [ 1., 1.73]])
        >>> dt = Delaunay(pts)
        >>> [myfrom,myto,nl] = VoronoiDelaunayGrid.create_links_from_triangulation(dt)
        >>> print myfrom, myto, nl # doctest: +SKIP
        [5 3 4 6 4 3 0 4 1 1 2 6] [3 4 5 5 6 0 4 1 0 2 4 2] 12
        """
    
        # Calculate how many links there will be and create the arrays.
        #
        # The number of links equals 3 times the number of triangles minus
        # half the number of shared links. Finding out the number of shared links
        # is easy: for every shared link, there is an entry in the tri.neighbors
        # array that is > -1 (indicating that the triangle has a neighbor opposite
        # a given vertex; in other words, two triangles are sharing an edge).
        #
        num_shared_links = numpy.count_nonzero(tri.neighbors>-1)
        num_links = 3*tri.nsimplex - num_shared_links/2
        link_fromnode = numpy.zeros(num_links, dtype=int)
        link_tonode = numpy.zeros(num_links, dtype=int)
        
        # Sweep through the list of triangles, assigning "from" and "to" nodes to
        # the list of links.
        #
        # The basic algorithm works as follows. For each triangle, we will add its
        # 3 edges as links. However, we have to make sure that each shared edge
        # is added only once. To do this, we keep track of whether or not each
        # triangle has been processed yet using a boolean array called "tridone".
        # When we look at a given triangle, we check each vertex in turn. If there
        # is no neighboring triangle opposite that vertex, then we need to add the
        # corresponding edge. If there is a neighboring triangle but we haven't
        # processed it yet, we also need to add the edge. If neither condition is
        # true, then this edge has already been added, so we skip it.
        link_id = 0
        tridone = numpy.zeros(tri.nsimplex, dtype=bool)    
        for t in range(tri.nsimplex):  # loop over triangles
            for i in range(0, 3):       # loop over vertices & neighbors
                if tri.neighbors[t,i] == -1 or not tridone[tri.neighbors[t,i]]:
                    link_fromnode[link_id] = tri.simplices[t,numpy.mod(i+1,3)]
                    link_tonode[link_id] = tri.simplices[t,numpy.mod(i+2,3)]
                    link_id += 1
            tridone[t] = True
        
        #save the results
        #self.link_fromnode = link_fromnode
        #self.link_tonode = link_tonode
        #self._num_links = num_links
    
        # Return the results
        return link_fromnode, link_tonode, num_links
    
    @staticmethod
    def is_valid_voronoi_ridge(vor, n):
        
        SUSPICIOUSLY_BIG = 40000000.0
        return vor.ridge_vertices[n][0]!=-1 and vor.ridge_vertices[n][1]!=-1 \
                and numpy.amax(numpy.abs(vor.vertices[vor.ridge_vertices[n]]))<SUSPICIOUSLY_BIG

    @staticmethod
    def create_links_and_faces_from_voronoi_diagram(vor):
        """
        From a Voronoi diagram object created by scipy.spatial.Voronoi(),
        builds and returns:
        1. Arrays of link "from" and "to" nodes
        2. Array of link IDs for each active link
        3. Array containing with of each face
        
        Parameters
        ----------
        vor : scipy.spatial.Voronoi
            Voronoi object initialized with the grid nodes.

        Returns
        -------
        out : tuple of ndarrays
            - link_fromnode = "from" node for each link (len=num_links)
            - link_tonode   = "to" node for each link (len=num_links)
            - active_links  = link ID for each active link (len=num_active_links)
            - face_width    = width of each face (len=num_active_links

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.grid import VoronoiDelaunayGrid
        >>> pts = np.array([[ 0., 0.],[  1., 0.],[  1.5, 0.87],[-0.5, 0.87],[ 0.5, 0.87],[  0., 1.73],[  1., 1.73]])
        >>> from scipy.spatial import Voronoi
        >>> vor = Voronoi(pts)
        >>> [fr,to,al,fw] = VoronoiDelaunayGrid.create_links_and_faces_from_voronoi_diagram(vor)
        >>> fr
        array([0, 0, 0, 1, 1, 3, 3, 6, 6, 6, 4, 4])
        >>> to
        array([3, 1, 4, 2, 4, 4, 5, 4, 2, 5, 2, 5])
        >>> al
        array([ 2,  4,  5,  7, 10, 11])
        >>> fw
        array([ 0.57669199,  0.57669199,  0.575973  ,  0.57836419,  0.575973  , 0.57836419])
        """
        # Each Voronoi "ridge" corresponds to a link. The Voronoi object has an
        # attribute ridge_points that contains the IDs of the nodes on either
        # side (including ridges that have one of their endpoints undefined).
        # So, we set the number of links equal to the number of ridges.
        num_links = len(vor.ridge_points)
        
        # Create the arrays for link from and to nodes
        link_fromnode = -numpy.ones(num_links, dtype=int)
        link_tonode = -numpy.ones(num_links, dtype=int)
        
        # Ridges along the perimeter of the grid will have one of their 
        # endpoints undefined. The endpoints of each ridge are contained in
        # vor.ridge_vertices, and an undefined vertex is flagged with -1.
        # Ridges with both vertices defined correspond to faces and active 
        # links, while ridges with an undefined vertex correspond to inactive
        # links. So, to find the number of active links, we subtract from the
        # total number of links the number of occurrences of an undefined
        # vertex.
        num_active_links = num_links \
                    - numpy.count_nonzero(numpy.array(vor.ridge_vertices)==-1)
        #print 'num_links=', num_links,'num_active_links=',num_active_links
        
        # Create arrays for active links and width of faces (which are Voronoi
        # ridges).
        active_links = -numpy.ones(num_active_links, dtype=int)
        face_width = -numpy.ones(num_active_links)
    
        # Loop through the list of ridges. For each ridge, there is a link, and
        # its "from" and "to" nodes are the associated "points". In addition, if
        # the ridge endpoints are defined, we have a face and an active link,
        # so we add them to our arrays as well.
        j = 0
        for i in range(num_links):
            link_fromnode[i] = vor.ridge_points[i,0]
            link_tonode[i] = vor.ridge_points[i,1]
            face_corner1 = vor.ridge_vertices[i][0]
            face_corner2 = vor.ridge_vertices[i][1]
            if VoronoiDelaunayGrid.is_valid_voronoi_ridge(vor, i):  # means it's a valid face
                dx = vor.vertices[face_corner2,0]-vor.vertices[face_corner1,0]
                dy = vor.vertices[face_corner2,1]-vor.vertices[face_corner1,1]
                face_width[j] = numpy.sqrt(dx*dx+dy*dy)
                if abs(face_width[j])>=40000.0:
                    print 'link',i,'from',link_fromnode[i],'to',link_tonode[i],'has face width',face_width[j]
                    print vor.ridge_vertices[i]
                    print vor.vertices[vor.ridge_vertices[i]]
                    from scipy.spatial import voronoi_plot_2d
                    voronoi_plot_2d(vor)
                assert face_width[j] < 40000., 'face width must be less than earth circumference!'
                active_links[j] = i
                j += 1
        #print 'active links:',active_links
        #print 'vor ridge points:',vor.ridge_points
        
        #save the data
        #self.link_fromnode = link_fromnode
        #self.link_tonode = link_tonode
        #self.active_link_ids = active_links
        #self._face_widths = face_width
        #self._num_faces = face_width.size
        #self._num_active_links = active_links.size
        
        return link_fromnode, link_tonode, active_links, face_width
        
    
    def reorient_links_upper_right(self):
        """
        Reorients links so that all point within the upper-right semi-circle.
        
        Notes
        -----
        "Upper right semi-circle" means that the angle of the link with respect
        to the vertical (measured clockwise) falls between -45 and +135. More
        precisely, if :math:`\theta' is the angle, :math:`-45 \ge \theta < 135`.
        For example, the link could point up and left as much as -45, but not -46.
        It could point down and right as much as 134.9999, but not 135. It will
        never point down and left, or up-but-mostly-left, or 
        right-but-mostly-down.
        
        Examples
        --------
        >>> from landlab.grid import HexModelGrid
        >>> hg = HexModelGrid(3, 2, 1., reorient_links=True)
        >>> hg.link_fromnode
        array([3, 3, 2, 0, 3, 1, 4, 5, 2, 0, 0, 1])
        >>> hg.link_tonode
        array([6, 5, 3, 3, 4, 3, 6, 6, 5, 2, 1, 4])
        """
        
        # Calculate the horizontal (dx) and vertical (dy) link offsets
        link_dx = self.node_x[self.link_tonode] - self.node_x[self.link_fromnode]
        link_dy = self.node_y[self.link_tonode] - self.node_y[self.link_fromnode]
        
        # Calculate the angle, clockwise, with respect to vertical, then rotate
        # by 45 degrees counter-clockwise (by adding pi/4)
        link_angle = numpy.arctan2(link_dx, link_dy) + numpy.pi/4
        
        # The range of values should be -180 to +180 degrees (but in radians).
        # It won't be after the above operation, because angles that were 
        # > 135 degrees will now have values > 180. To correct this, we subtract
        # 360 (i.e., 2 pi radians) from those that are > 180 (i.e., > pi radians).
        link_angle -= 2*numpy.pi*(link_angle>=numpy.pi)
        
        # Find locations where the angle is negative; these are the ones we
        # want to flip
        (flip_locs, ) = numpy.where(link_angle<0.)
        
        # If there are any flip locations, proceed to switch their fromnodes and
        # tonodes; otherwise, we're done
        if len(flip_locs)>0:
            
            # Temporarily story the fromnode for these
            fromnode_temp = self.link_fromnode[flip_locs]
            
            # The fromnodes now become the tonodes, and vice versa
            self.link_fromnode[flip_locs] = self.link_tonode[flip_locs]
            self.link_tonode[flip_locs] = fromnode_temp
            
    def create_patches_from_delaunay_diagram(self, pts, vor, nodata=-1):
        """
        Uses a delaunay diagram drawn from the provided points to
        generate an array of patches and patch-node-link connectivity.
        Returns ...
        DEJH, 10/3/14
        """
        from scipy.spatial import Delaunay
        tri = Delaunay(pts)
        assert numpy.array_equal(tri.points, vor.points)
        
        if nodata==-1:
            pass
        elif nodata=='bad_value':
            nodata=BAD_INDEX_VALUE
        elif nodata=='nan':
            nodata=numpy.nan
        else:
            raise ValueError('Do not recognise nodata value!')
        
        self._patch_nodes = tri.simplices
        self._number_of_patches = tri.simplices.shape[0]
        max_dimension = 0
        #need to build a squared off, masked array of the node_patches
        #the max number of patches for a node in the grid is the max sides of
        #the side-iest voronoi region.
        for i in xrange(len(vor.regions)):
            if len(vor.regions[i])>max_dimension:
                max_dimension=len(vor.regions[i])
        _node_patches = numpy.empty((self.number_of_nodes, max_dimension), dtype=int)
        _node_patches.fill(nodata)
        for i in xrange(self.number_of_nodes):
            if not self.is_boundary(i, boundary_flag=4): #don't include closed nodes
                patches_with_node = numpy.argwhere(numpy.equal(self._patch_nodes,i))[:,0]
                _node_patches[i,:patches_with_node.size] = patches_with_node[:]
        #mask it
        self._node_patches = numpy.ma.array(_node_patches, mask=numpy.equal(_node_patches, -1))



#################################################################
##
##  Potential Evapotranspiration Component calculate spatially distributed
##  potential evapotranspiration based on input radiation factor (spatial
##  distribution of incoming radiation) using chosen method such as a
##  constant or Priestly Taylor.
##
##  Sai Nudurupati and Erkan Istanbulluoglu - 16May2014
##  Added Priestly Taylor on 09Jul2014
#################################################################

from landlab import Component

import numpy as np

_VALID_METHODS = set(['Constant', 'PriestlyTaylor', 'MeasuredRadiationPT',
                            'Cosine'])

def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class PotentialEvapotranspiration( Component ):
    """
    Landlab component that calculates Potential Evapotranspiration.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.PET.potential_evapotranspiration_field import PotentialEvapotranspiration

    >>> grid = RasterModelGrid(5, 4, 1.e4)
    >>> PET = PotentialEvapotranspiration(grid)
    >>> PET.name
    'Potential Evapotranspiration'
    """
    _name = 'Potential Evapotranspiration'

    _input_var_names = set([
        'RadiationFactor',
    ])

    _output_var_names = set([
        'TotalShortWaveRadiation',
        'NetShortWaveRadiation',
        'NetLongWaveRadiation',
        'NetRadiation',
        'PotentialEvapotranspiration',
    ])

    _var_units = {
        'PotentialEvapotranspiration' : 'mm',
        'TotalShortWaveRadiation' : 'W/m^2',
        'NetShortWaveRadiation' : 'W/m^2',
        'NetLongWaveRadiation' : 'W/m^2',
        'NetRadiation' : 'W/m^2',
        'RadiationFactor' : 'None',
    }

    def __init__(self, grid, **kwds):
        self._method = kwds.pop('method', 'Constant')
        # For Priestly Taylor
        self._alpha = kwds.pop('PriestlyTaylorConstant', 1.26)
        self._a = kwds.pop('Albedo', 0.6)
        self._pwhv = kwds.pop('LatentHeatofVaporization', 28.34)
        self._y = kwds.pop('PsychometricConstant', 0.066)
        self._sigma = kwds.pop('StefanBoltzmannConstant', 0.0000000567)
        self._Gsc = kwds.pop('SolarConstant', 1366.67)
        self._phi = (3.14/180)*(kwds.pop('Latitude', 34.0))
        self._z = kwds.pop('ElevationofMeasurement', 300)
        self._Krs = kwds.pop('AdjustmentCoefficient', 0.18)
        self._LT = kwds.pop('LT', 0.)
        self._ND = kwds.pop('ND', 365.)
        self._TmaxF_mean = kwds.pop('MeanTmaxF', 12.)
        self._DeltaD= kwds.pop('DeltaD', 5.)

        assert_method_is_valid(self._method)

        super(PotentialEvapotranspiration, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if not name in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if not name in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._cell_values = self.grid['cell']

    def update(self, current_time, **kwds):

        if self._method == 'Constant':
            self._PET_value = kwds.pop('ConstantPotentialEvapotranspiration', 12.)
        elif self._method == 'PriestlyTaylor':
            Tmin = kwds.pop('Tmin',0.0)
            Tmax = kwds.pop('Tmax',1.0)
            Tavg = kwds.pop('Tavg',0.5)
            self._PET_value = self.PriestlyTaylor( current_time,Tmax, Tmin, Tavg )
            self._cell_values['TotalShortWaveRadiation'] = self._Rs * \
                                        self._cell_values['RadiationFactor']
            self._cell_values['NetShortWaveRadiation'] = self._Rns * \
                                        self._cell_values['RadiationFactor']
            self._cell_values['NetLongWaveRadiation'] = self._Rnl * \
                                        self._cell_values['RadiationFactor']
            self._cell_values['NetRadiation'] = self._Rn * \
                                        self._cell_values['RadiationFactor']
        elif self._method == 'MeasuredRadiationPT':
            Tavg = kwds.pop('Tavg',0.5)
            Robs = kwds.pop('Radiation', 350.)
            self._PET_value = self.MeasuredRadPT( Tavg, (1-self._a)*Robs )
        elif self._method == 'Cosine':
            self._J = np.floor( (current_time - np.floor( current_time)) * 365.)
            self._PET_value = max((self._TmaxF_mean + self._DeltaD/2. * np.cos((2*np.pi) *
                            (self._J - self._LT - self._ND/2)/self._ND)), 0.0)

        self._PET = self._PET_value * self._cell_values['RadiationFactor']
        self._cell_values['PotentialEvapotranspiration'] = self._PET


    def PriestlyTaylor(self, current_time, Tmax, Tmin, Tavg):

        """
            Julian Day - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 25, (52)
        """
        self._J = np.floor( (current_time - np.floor( current_time)) * 365 )

        """
            Saturation Vapor Pressure - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 6, (37)
        """
        self._es = 0.6108*np.exp((17.27*Tavg)/(237.7+Tavg))
        """
            Actual Vapor Pressure - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 8, (38)
        """
        self._ea = 0.6108*np.exp((17.27*Tmin)/(237.7+Tmin))
        """
            Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 5, (36)
        """
        self._delta = (4098.0*self._es)/((237.3+Tavg)**2.0)
        """
            Solar Declination Angle - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 24,(51)
        """
        self._sdecl = 0.409*np.sin((((2.0*3.14)/365.0)*self._J)-1.39)
        """
            Inverse Relative Distance Factor - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 23,(50)
        """
        self._dr = 1 + (0.033*np.cos((2.0*3.14/365.0)*self._J))
        """
            To calculate ws - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 29,(61)
        """
        self._x = 1.0-(((np.tan(self._phi))**2.0)*(np.tan(self._sdecl)**2.0))
        if self._x <= 0:
            self._x = 0.00001;

        """
            Sunset Hour Angle - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 28,(60)
        """
        self._ws = (3.14/2.0)-np.arctan((-1*np.tan(self._phi)*np.tan(self._sdecl))/(self._x**2.0))
        """
            Extraterrestrial radmodel.docx - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 21, (48)
        """
        self._Ra = 11.57*(24.0/3.14)*4.92*self._dr*((self._ws*np.sin(self._phi)*np.sin(self._sdecl))+ \
                        (np.cos(self._phi)*np.cos(self._sdecl)*(np.sin(self._ws))))
        """
            Clear-sky Solar Radiation - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 19, (47)
        """
        self._Rso = (0.75+((2.0*(10**(-5.0)))*self._z))*self._Ra
        self._Rs = min(self._Krs*self._Ra*np.sqrt(Tmax-Tmin), self._Rso)
        """
            Net Short Wave Radiation - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 16, (43)
        """
        self._Rns = self._Rs*(1-self._a)

        """
            Relative Cloudiness - ASCE-EWRI Task Committee Report, Jan-2005 - Page 20,35
        """
        if self._Rso > 0:
            self._u = self._Rs/self._Rso
        else:
            self._u = 0

        if self._u < 0.3:
            self._u = 0.3
        elif self._u > 1:
            self._u = 1.0

        """
            Cloudiness Function - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 18, (45)
        """
        self._fcd = (1.35*self._u)-0.35
        """
            Net Long Wave Radiation - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 17, (44)
        """
        self._Rnl = self._sigma*self._fcd*(0.34-(0.14*np.sqrt(self._ea))* \
                        (((Tmax+273.16)**4.0+ \
                        (Tmin+273.16)**4.0)/2.0))

        """
            Net Radiation - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 15, (42)
        """
        self._Rn = self._Rns - self._Rnl

        self._ETp = max(self._alpha*(self._delta/(self._delta+self._y) \
                        )*(self._Rn/self._pwhv), 0)

        return( self._ETp )


    def MeasuredRadPT( self, Tavg, Rnobs ):

        """
            Saturation Vapor Pressure - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 6, (37)
        """
        self._es = 0.6108*np.exp((17.27*Tavg)/(237.7+Tavg))
        """
            Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 5, (36)
        """
        self._delta = (4098.0*self._es)/((237.3+Tavg)**2.0)

        self._ETp = max(self._alpha*(self._delta/(self._delta+self._y) \
                        )*( Rnobs/self._pwhv), 0)

        return( self._ETp )



"""
simple script to run speed tests of various functions in model grid
"""

import landlab as ll
import time


def main():
    mg = ll.RasterModelGrid(20, 30, 1.0)

    nt = 1000

    s = mg.zeros(centering='node')
    g = mg.zeros(centering='active_link')
    divg = mg.zeros(centering='node')

    start_time = time.time()

    for i in range(nt):
        
        g = mg.calculate_gradients_at_active_links(s, g)
        
    time1 = time.time()

    for i in range(nt):
        
        g = mg._calculate_gradients_at_active_links_slow(s, g)
        
    time2 = time.time()

    for i in range(nt):
        
        divg = mg.calculate_flux_divergence_at_nodes(g, divg)
        
    time3 = time.time()

    for i in range(nt):
        
        divg = mg.calculate_flux_divergence_at_nodes_slow(g, divg)

    time4 = time.time()
      
    for i in range(nt):
        
        divg = mg.calculate_flux_divergence_at_active_cells(g)
        
    time5 = time.time()

    for i in range(nt):
        
        divg = mg.calculate_flux_divergence_at_active_cells_slow(g)

    time6 = time.time()
      
    print('Elapsed time with fast gradient algo: '+str(time1-start_time))
    print('Elapsed time with slow gradient algo: '+str(time2-time1))
    print('Elapsed time with fast node-divergence algo: '+str(time3-time2))
    print('Elapsed time with slow node-divergence algo: '+str(time4-time3))
    print('Elapsed time with fast activecell-divergence algo: '+str(time5-time4))
    print('Elapsed time with slow activecell-divergence algo: '+str(time6-time5))


if __name__ == '__main__':
    main()



#import numpy as np

from landlab import Implements, ImplementsOrRaise
from landlab.framework.interfaces import BmiBase, BmiNoGrid


#@ImplementsOrRaise(BmiBase)
@Implements(BmiBase)
class Sample1(object):
    """
    A sample component.
    """
    __implements__ = (BmiBase, BmiNoGrid, )

    _input_var_names = [
        'air__temperature',
        'surface__elevation',
    ]
    _output_var_names = [
        'deposition__rate',
    ]

    model_name = 'Sample 1'
    author_name = 'Eric Hutton'
    version = '0.1'
    time_units = 's'
    time_step_type = 'fixed'
    step_method = 'explicit'
    grid_type = 'none'

    _vars = {
        'deposition__rate': [1.]
    }

    def initialize(self, name):
        pass

    def update(self):
        pass

    def finalize(self):
        pass

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_var_rank(self, name):
        return 0

    def get_start_time(self):
        return 0.

    def get_current_time(self):
        return 0.

    def get_end_time(self):
        return 100.

    def get_time_step(self):
        return 1.

    def get_var_type(self, name):
        return 'float64'

    def get_var_units(self, name):
        return 'm'

    def set_value(self, name, value):
        pass

    def get_value(self, name):
        return self._vars[name]


@ImplementsOrRaise(BmiBase)
class Sample2(object):
    """
    A sample component.
    """
    __implements__ = (BmiBase, BmiNoGrid, )

    _input_var_names = [
        'deposition__rate',
    ]
    _output_var_names = [
        'air__temperature',
        'surface__elevation',
    ]

    model_name = 'Sample 2'
    author_name = 'Eric Hutton'
    version = '0.1'
    time_units = 's'
    time_step_type = 'fixed'
    step_method = 'explicit'
    grid_type = 'none'

    _vars = {
        'air__temperature': [1.],
        'surface__elevation': [1.],
    }

    def initialize(self, name):
        pass

    def update(self):
        pass

    def finalize(self):
        pass

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_var_rank(self, name):
        return 0

    def get_start_time(self):
        return 0.

    def get_current_time(self):
        return 0.

    def get_end_time(self):
        return 100.

    def get_time_step(self):
        return 1.

    def get_var_type(self, name):
        return 'float64'

    def get_var_units(self, name):
        return 'm'

    def get_value(self, name):
        return self._vars[name]

    def set_value(self, name, value):
        pass



""" generate_uniform_precip.py 
 This component generates rainfall  
 events based on statistical distributions.

 No particular units must be used, but it was
 written with the storm units in hours (hr)
and depth units in millimeters (mm)


 Written by Jordan Adams, 2013.
"""

import os
import numpy as np 
import random
from landlab import Component,ModelParameterDictionary
from landlab.core.model_parameter_dictionary import MissingKeyError

_DEFAULT_INPUT_FILE = os.path.join(os.path.dirname(__file__),
                                  'preciptest.in')

class PrecipitationDistribution(Component):
    """Landlab component that generates precipitation events
    using the rectangular Poisson pulse model described in
    Eagleson (1978).
    
    This component can generate a random storm duration, interstorm
    duration, precipitation intensity or storm depth from a Poisson 
    distribution when given a mean value.
    
    Default input file is named 'preciptest.in' and can be found in
    the landlab.components.uniform_precip folder.
    
        Inputs
        ------
        input_file : Contains necessary inputs. If not given, default input file is used.
            - MEAN_STORM: (type : float) the mean storm duration if not provided in initialization
            - MEAN_DEPTH: (type : float) the mean storm depth if not provided in initizalization
            - MEAN_INTERSTORM: (type : float) the mean interstorm duration if not provided in initialization
            - RUN_TIME: (type : float) total model run time if not provided in initialization
            - DELTA_T: (type : int) external time step increment if not provided in initialization 
                (it is not obligtory to provide a DELTA_T)
            
    So, without an input file (selecting the default), we can call this component like...
    
    >>> from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution
    >>> precip = PrecipitationDistribution()
        
    To use hard-coded values for mean storm, mean interstorm, mean depth, model run time and delta t...
    Say we use 1.5 for mean storm, 15 for mean interstorm, 0.5 for mean depth, 100 for model run time and 1 for delta t...
    
    >>> precip = PrecipitationDistribution(input_file=None,
    ...     mean_storm=1.5, mean_interstorm=15.0, mean_storm_depth=0.5,
    ...     total_t=100.0, delta_t=1)
    """

    def __init__(self, input_file=None, mean_storm=None, mean_interstorm=None, mean_storm_depth=None, total_t=None, delta_t=None):
        """ This reads in information from the input_file (either default or user
            assigned, and creates an instantaneous storm event drawn from the Poisson distribution
        """
        
        # First we create an instance of the Model Parameter Dictionary
        MPD = ModelParameterDictionary()
        
        # If no input_file is given,the default file is used
        if input_file is None:
            input_file = _DEFAULT_INPUT_FILE
            
        # This reads in the file information    
        MPD.read_from_file(input_file)       
        
        # And now we set our different parameters 
        # using the model parameter dictionary...
        if mean_storm == None:
            self.mean_storm = MPD.read_float( 'MEAN_STORM')
        else:
            self.mean_storm = mean_storm
        
        if mean_interstorm == None:
            self.mean_interstorm = MPD.read_float( 'MEAN_INTERSTORM')
        else:
            self.mean_interstorm =mean_interstorm
            
        if mean_storm_depth== None:
            self.mean_storm_depth = MPD.read_float( 'MEAN_DEPTH')
        else:
            self.mean_storm_depth =mean_storm_depth
            
        if total_t== None:
            self.run_time = MPD.read_float( 'RUN_TIME')
        else:
            self.run_time =total_t
            
        if delta_t== None:
            if input_file != _DEFAULT_INPUT_FILE:
                try:
                    self.delta_t = MPD.read_float( 'DELTA_T')
                except MissingKeyError:
                    self.delta_t = None
            else:
                self.delta_t = None
        else:
            self.delta_t =delta_t
            
        # Mean_intensity is not set by the MPD, but can be drawn from 
        # the mean storm depth and mean storm duration.
        self.mean_intensity = self.mean_storm_depth / self.mean_storm

        # If a time series is created later, this blank list will be used.
        self.storm_time_series =[]

        # Given the mean values assigned above using either the model
        # parameter dictionary or the init function, we can call the
        # different methods to assign values from the Poisson distribution.
        
        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()
        self._elapsed_time = 0.


    def update(self):
        """If new values for storm duration, interstorm duration, storm depth
        and intensity are needed, this method can be used to update those values
        one time. 
        
        >>> from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution
        >>> PD = PrecipitationDistribution()
        >>> PD.update()
            
        Additionally, if we wanted to update several times, a loop could be
        utilized to accomplish this. Say we want 5 storm_durations; this
        pseudo-code represents a way to accomplish this...
        
        >>> PD = PrecipitationDistribution()
        >>> storm_duration_list=[]
        >>> i = 0
        >>> while i < 4:
        ...     storm_duration_list.append(PD.storm_duration)
        ...     PD.update()
        ...     i+=1        
        """
        
        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()
        
    def get_precipitation_event_duration(self):
        """This method is the storm generator.
    
    This method has one argument: the mean_storm parameter.
    (In Eagleson (1978), this parameter was called Tr.)
    
    It finds a random storm_duration value
    based on the poisson distribution about the mean.
    This is accomplished using the expovariate function
    from the "random" standard library.
    Additionally, it is rounded to contain 4 significant figures, 
    for neatness. 
    
    The if-else statement is very important here. Values of 0
    can exist in the Poission distribution, but it does not make
    sense to have 0 duration storms, so to avoid that,
    we return a storm duration IF it is greater than 0,
    otherwise, recursion is employed to re-call the storm
    generator function and get a new value.
    
    :returns: storm_duration as a float
    """
        storm = round(random.expovariate(1/self.mean_storm),2)
        while storm == 0: 
            storm = round(random.expovariate(1/self.mean_storm),2)
        self.storm_duration = storm
        return self.storm_duration

    
    
    def get_interstorm_event_duration(self):
        """ This method is the interstorm duration generator
    
    This method takes one argument, the mean_interstorm parameter.
    (In Eagleson (1978), this parameter was called Tb.)
    
    This method is modeled identically to get_precipitation_event_duration()
    
    This method finds a random value for interstorm_duration
    based on the poisson distribution about the mean.
    This is accomplished using the expovariate function
    from the "random" standard library.
    Additionally, it is rounded to contain 4 significant figures, for neatness. 
    
    The if-else statement is very important here. Values of 0
    can exist in the Poission distribution, but it does not make
    sense to have 0 hour interstorm durations.
    To avoid 0 hour interstorm durations, we return a 
    interstorm duration IF it is greater than 0,
    otherwise, recursion is employed to re-call the interstorm
    duration generator function and get a new value.
    
    :returns: interstorm_duration as a float"""
    
        interstorm = round(random.expovariate(1/self.mean_interstorm),2)
        while interstorm == 0:
            interstorm = round(random.expovariate(1/self.mean_interstorm),2)
        self.interstorm_duration = interstorm
        return self.interstorm_duration
     
                            
    def get_storm_depth(self):
        """  This method is the storm depth generator.
    Storm depth is used to generate a realistic 
    intensity for different storm events.
    
    (In Eagleson (1978) this parameter was called "h")
    
    This method requires storm_duration, mean_storm duration
    and the mean_storm_depth. Storm_duration is generated through
    the initialize() or update() method. mean_storm and mean_storm_depth
    are read in using the ModelParameterDictionary.
    
    Numpy has a random number generator to get values
    from a given Gamma distribution. It takes two arguments,
    alpha (or the shape parameter), which is the generated over the mean event
    and beta (or the scale parameter), which is the mean value
    These are all arguments in the function, which returns storm depth.
    
    :returns: storm_depth as a float
    """
        
        shape_parameter = (self.storm_duration/self.mean_storm)
        scale_parameter = (self.mean_storm_depth)
        self.storm_depth = np.random.gamma(shape_parameter, scale_parameter)
        return self.storm_depth

    
    def get_storm_intensity(self):
        """   This method draws storm intensity out of the storm depth
    generated by get_storm_depth. 
    
    This method requires the storm_depth and storm_duration
    and is the same as the parameter ("i") in Eagleson (1978), but instead of
    being drawn from Poission, this is drawn from the Gamma distribution
    of ("h"), as h = i*Tr. 
    
    :returns: storm_intensity as a float
                            
                            """
        self.intensity = self.storm_depth / self.storm_duration
        return self.intensity


    def get_storm_time_series(self):
        """
        This method creates a time series of storms based on storm_duration, and
        interstorm_duration. From these values it will calculate a complete
        time series.
        
        The storm_time_series returned by this method is made up of sublists, each comprising of three
        sub-parts (e.g. [[x,y,z], [a,b,c]]) where x and a are the beginning times of a precipitation 
        event, y and b are the ending times of the precipitation event and z and c represent the
        average intensity (mm/hr) of the storm lasting from x to y and a to be, respectively. 
        :returns: array containing several sub-arrays of events [start, finish, intensity]
        
        
        There is a helper and an iterator. Open to suggestions on how to make this 
        sleeker. helper keeps track of storm events so that we can properly adjust the time for the 
        storm_time_series list. 
        
        Storm iterator makes sure that the run goes through to the specified time, either read in as "run_time" by
        the ModelParameterDictionary or specified the fire time this method is called."""
        
        storm = self.get_precipitation_event_duration()
        self.get_storm_depth()
        intensity = self.get_storm_intensity()
        self.storm_time_series.append([0, storm, intensity])

        storm_helper = storm
        storm_iterator = storm
        while storm_iterator <= self.run_time:
            next_storm_start  = storm_helper + round(self.get_interstorm_event_duration(),2)
            next_storm_end = next_storm_start + round(self.get_precipitation_event_duration(),2)
            intensity = round(self.get_storm_intensity(),2)
            self.get_storm_depth()
            self.storm_time_series.append([next_storm_start, next_storm_end, intensity])
            storm_iterator = storm_helper
            storm_helper = next_storm_end
            storm_iterator = storm_helper
        return self.storm_time_series
            
    def yield_storm_interstorm_duration_intensity(self, subdivide_interstorms=False):
        """
        This method is intended to be equivalent to get_storm_time_series,
        but instead offers a generator functionality. This will be useful in
        cases where the whole sequence of storms and interstorms doesn't need
        to be stored, where we can save memory this way.
        
        The method keeps track of the DELTA_T such that if a storm needs to be
        generated longer than this supplied model timestep, the generator will
        return the storm in "chunks", until there is no more storm duration.
        e.g.,
        storm of intensity 1. is 4.5 long, the DELTA_T is 2., the generator 
        yields (2.,1.) -> (2.,1.) -> (0.5,1.) -> ...
        
        If DELTA_T is None or not supplied, no subdivision occurs.
        
        Once a storm has been generated, this method will follow it with the
        next interstorm, yielded as (interstorm_duration, 0.). Note that the
        interstorm will NOT be subdivided according to DELTA_T unless you set
        the flag *subdivide_interstorms* to True.
        
        The method will keep yielding until it reaches the RUN_TIME, where it
        will terminate.
        
        YIELDS:
            - a tuple, (interval_duration, rainfall_rate_in_interval)
            
        One recommended procedure is to instantiate the generator, then call
        instance.next() repeatedly to get the sequence.
            
        Added DEJH, Dec 2014
        """
        delta_t = self.delta_t
        if delta_t == None:
            assert subdivide_interstorms == False, 'You specified you wanted storm subdivision, but did not provide a DELTA_T to allow this!'
        self._elapsed_time = 0.
        while self._elapsed_time<self.run_time:
            storm_duration = self.get_precipitation_event_duration()
            step_time=0.
            self.get_storm_depth()
            intensity = self.get_storm_intensity() #this is a VELOCITY, i.e., a rainfall rate
            if self._elapsed_time+storm_duration>self.run_time:
                storm_duration = self.run_time-self._elapsed_time
            while delta_t!=None and storm_duration-step_time>delta_t:
                yield (delta_t, intensity)
                step_time+=delta_t
            yield (storm_duration-step_time, intensity)
            self._elapsed_time += storm_duration
            
            interstorm_duration = self.get_interstorm_event_duration()
            if self._elapsed_time+interstorm_duration>self.run_time:
                interstorm_duration = self.run_time-self._elapsed_time
            if subdivide_interstorms:
                step_time=0.
                while interstorm_duration-step_time>delta_t:
                    yield (delta_t, 0.)
                    step_time += delta_t
                yield (interstorm_duration-step_time, 0.)
            else:
                yield (interstorm_duration, 0.)
            self._elapsed_time += interstorm_duration
    
    @property
    def elapsed_time(self):
        """
        Return the elapsed time recorded by the module.
        This will be particularly useful in the midst of a yield loop.
        """
        return self._elapsed_time



import numpy as np
from landlab import ModelParameterDictionary
from time import sleep
from scipy import weave
from scipy.weave.build_tools import CompileError
from landlab.utils import structured_grid as sgrid
import pylab

from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.field.scalar_data_fields import FieldError

class TransportLimitedEroder(object):
    """
    This component implements transport limited erosion for a landscape in which
    flow directions are fully convergent. i.e., all nodes in the landscape have
    a single, uniquely defined downstream node.
    
    The module can in principle take multiple transport laws, but at the moment
    only Meyer-Peter Muller (MPM) is implemented.
    
    There is as yet no explicit stabilization check on the timestep. If your
    run destabilizes, try reducing dt.
    See, e.g., ./examples/simple_sp_driver.py
    
    Assumes grid does not deform or change during run.
    
    Note it is vital to ensure all your units match. t is assumed to be in 
    years. Any length value is assumed to be in meters (including both dx
    and the uplift rate...!)
    
    DEJH Sept 14.
    Currently only runs on a raster grid
    """
    
    def __init__(self, grid, params):
        self.initialize(grid, params)
            
#This draws attention to a potential problem. It will be easy to have modules update z, but because noone "owns" the data, to forget to also update dz/dx...
#How about a built in grid utility that updates "derived" data (i.e., using only grid fns, e.g., slope, curvature) at the end of any given tstep loop?
#Or an explicit flagging system for all variables in the modelfield indicating if they have been updated this timestep. (Currently implemented)
#Or wipe the existance of any derived grid data at the end of a timestep entirely, so modules find they don't have it next timestep.
        
    def initialize(self, grid, params_file):
        '''
        params_file is the name of the text file containing the parameters 
        needed for this stream power component.
        
        ***Parameters for input file***
        OBLIGATORY:
            * Qc -> String. Controls how to set the carrying capacity.
                Either 'MPM', or a string giving the name of the model field
                where capacity values are stored on nodes.
                At the moment, only 'MPM' is permitted as a way to set the 
                capacity automatically, but expansion would be trivial.
                If 'from_array', the module will attempt to set the capacity
                Note capacities must be specified as volume flux.
            * 
            
            ...Then, assuming you set Qc=='MPM':
            * b_sp, c_sp -> Floats. These are the powers on discharge and 
                drainage area in the equations used to control channel width and 
                basin hydrology, respectively:
                        W = k_w * Q**b_sp
                        Q = k_Q * A**c_sp
                These parameters are used to constrain flow depth, and may be
                omitted if use_W or use_Q are set.
            *k_Q, k_w, mannings_n -> floats. These are the prefactors on the 
                basin hydrology and channel width-discharge relations, and n
                from the Manning's equation, respectively. These are 
                needed to allow calculation of shear stresses and hence carrying
                capacities from the local slope and drainage area alone.
                The equation for depth used to derive shear stress and hence
                carrying capacity contains a prefactor:
                    mannings_n*(k_Q**(1-b)/K_w)**0.6 
                (so shear = fluid_density*g*depth_equation_prefactor*A**(0.6*c*(1-b)*S**0.7 !)
                Don't know what to set these values to? k_w=0.002, k_Q=1.e-9,
                mannings_n=0.03 give vaguely plausible numbers (e.g., for a
                drainage area ~400km2, like Boulder Creek at Boulder,
                => depth~2.5m, width~35m, shear stress ~O(1000Pa)).
            *Dchar -> float.  The characteristic grain diameter in meters
                (==D50 in most cases) used to calculate Shields numbers
                in the channel. If you want to define Dchar values at each node,
                don't set, and use the Dchar_if_used argument in erode() 
                instead.
            
        OPTIONS:
            *rock_density -> in kg/m3 (defaults to 2700)
            *sediment_density -> in kg/m3 (defaults to 2700)
            *fluid_density -> in most cases water density, in kg/m3 (defaults to 1000)
            *g -> acceleration due to gravity, in m/s**2 (defaults to 9.81)
            
            *threshold_shields -> +ve float; the threshold taustar_crit. 
                Defaults to 0.047, or if 'slope_sensitive_threshold' is set True,
                becomes a weak function of local slope following Lamb et al 
                (2008):
                    threshold_shields=0.15*S**0.25
            *slope_sensitive_threshold -> bool, defaults to 'False'.
                If true, threshold_shields is set according to the Lamb
                equation. An exception will be raised if threshold_shields is 
                also set.
            *Parker_epsilon -> float, defaults to 0.4. This is Parker's (1978)
                epsilon, which is used in the relation
                    tau - tauc = tau * (epsilon/(epsilon+1))
                The 0.4 default is appropriate for coarse (gravelly) channels.
                The value approaches infinity as the river banks become more
                cohesive.
            *dt -> +ve float. If set, this is the fixed timestep for this
                component. Can be overridden easily as a parameter in erode(). 
                If not set (default), this parameter MUST be set in erode().
            *use_W -> Bool; if True, component will look for node-centered data
                describing channel width in grid.at_node['channel_width'], and 
                use it to implement incision ~ stream power per unit width. 
                Defaults to False.
            *use_Q -> Bool. Overrides the basin hydrology relation, using an
                local water discharge value assumed already calculated and
                stored in grid.at_node['discharge'].
            *C_MPM -> float. Defaults to 1. Allows tuning of the MPM prefactor,
                which is calculated as
                    Qc = 8.*C_MPM*(taustar - taustarcrit)**1.5
                In almost all cases, tuning depth_equation_prefactor' is 
                preferred to tuning this parameter.
            *return_capacity -> bool (default False). NOT YET IMPLEMENTED.
                If True, this component
                will save the calculated capacity in the field 
                'fluvial_sediment_transport_capacity'. (Requires some additional
                math, so is suppressed for speed by default).
            
        '''
        self.grid = grid
        self.link_S_with_trailing_blank = np.zeros(grid.number_of_links+1) #needs to be filled with values in execution
        self.count_active_links = np.zeros_like(self.link_S_with_trailing_blank, dtype=int)
        self.count_active_links[:-1] = 1
        inputs = ModelParameterDictionary(params_file)
        try:
            self.g = inputs.read_float('g')
        except MissingKeyError:
            self.g = 9.81
        try:
            self.rock_density = inputs.read_float('rock_density')
        except MissingKeyError:
            self.rock_density = 2700.
        try:
            self.sed_density = inputs.read_float('sediment_density')
        except MissingKeyError:
            self.sed_density = 2700.
        try:
            self.fluid_density = inputs.read_float('fluid_density')
        except MissingKeyError:
            self.fluid_density = 1000.
        self.rho_g = self.fluid_density * self.g
        
        try:
            self.Qc = inputs.read_string('Qc')
        except MissingKeyError:
            raise MissingKeyError("Qc must be 'MPM' or a grid field name!")
        else:
            if self.Qc=='MPM':
                self.calc_cap_flag = True
            else:
                self.calc_cap_flag = False
                
        try:
            self.lamb_flag = inputs.read_bool('slope_sensitive_threshold')
        except:
            self.lamb_flag = False
        try:
            self.shields_crit = inputs.read_float('threshold_shields')
            self.set_threshold = True #flag for sed_flux_dep_incision to see if the threshold was manually set.
            print "Found a threshold to use: ", self.shields_crit
            assert self.lamb_flag == False
        except MissingKeyError:
            if not self.lamb_flag:
                self.shields_crit = 0.047
            self.set_threshold = False
        try:
            self.tstep = inputs.read_float('dt')
        except MissingKeyError:
            pass
        try:
            self.use_W = inputs.read_bool('use_W')
        except MissingKeyError:
            self.use_W = False
        try:
            self.use_Q = inputs.read_bool('use_Q')
        except MissingKeyError:
            self.use_Q = False
        try:
            self.return_capacity = inputs.read_bool('return_capacity')
        except MissingKeyError:
            self.return_capacity = False
            
        try:
            self._b = inputs.read_float('b_sp')
        except MissingKeyError:
            if self.use_W:
                self._b = 0.
            else:
                if self.calc_cap_flag:
                    raise NameError('b was not set')
        try:
            self._c = inputs.read_float('c_sp')
        except MissingKeyError:
            if self.use_Q:
                self._c = 1.
            else:
                if self.calc_cap_flag:
                    raise NameError('c was not set')
        try:
            self.Dchar_in = inputs.read_float('Dchar')
        except MissingKeyError:
            pass
            
        #assume Manning's equation to set the power on A for shear stress:
        self.shear_area_power = 0.6*self._c*(1.-self._b)
        
        self.k_Q = inputs.read_float('k_Q')
        self.k_w = inputs.read_float('k_w')
        mannings_n = inputs.read_float('mannings_n')
        if mannings_n<0. or mannings_n>0.2:
            print "***STOP. LOOK. THINK. You appear to have set Manning's n outside it's typical range. Did you mean it? Proceeding...***"
            sleep(2)
        self.depth_prefactor = self.rho_g*mannings_n*(self.k_Q**(1.-self._b)/self.k_w)**0.6
        ##Note the depth_prefactor we store already holds rho*g   
        try:
            epsilon = inputs.read_float('Parker_epsilon')
        except MissingKeyError:
            epsilon = 0.4

        try:
            self.C_MPM = inputs.read_float('C_MPM')
        except MissingKeyError:
            self.C_MPM = 1.
        try:
            self.shields_prefactor = 1./((self.sed_density-self.fluid_density)*self.g*self.Dchar_in)
            self.MPM_prefactor = 8.*self.C_MPM*np.sqrt(self.relative_weight*self.Dchar_in*self.Dchar_in*self.Dchar_in)
            self.MPM_prefactor_alt = 4.*self.g**(-2./3.)/self.excess_SG/self.fluid_density/self.sed_density
        except AttributeError:
            #have to set these manually as needed
            self.shields_prefactor_noD = 1./((self.sed_density-self.fluid_density)*self.g)
        self.diffusivity_prefactor = 8.*np.sqrt(8.*self.g)/(self.sed_density/self.fluid_density-1.)*(epsilon/(epsilon+1.))**1.5*mannings_n**(5./6.)*self.k_w**-0.9*self.k_Q**(0.9*(1.-self._b)) #...this is multiplied by A**c(1-0.1*(1-b))
        #we consciously skip out a factor of S**0.05-->1. in the diffusion prefactor, to avoid delinearizing the diffusion. Only a possible problem at tiny S (20% error @S==0.01; 37% error @S==10**-4)
        #we could include this as a static adjustment in the actual looping code (i.e., just multiply by S**0.05, and don't work with it as part of the problem)
        #in reality, Manning's n changes downstream too, so... whatever
        self.diffusivity_power_on_A = 0.9*self._c*(1.-self._b) #i.e., q/D**(1/6)

        self.cell_areas = np.empty(grid.number_of_nodes)
        self.cell_areas.fill(np.mean(grid.cell_areas))
        self.cell_areas[grid.cell_node] = grid.cell_areas
        self.dx2 = grid.node_spacing_horizontal**2
        self.dy2 = grid.node_spacing_vertical**2
        self.bad_neighbor_mask = np.equal(grid.get_neighbor_list(bad_index=-1),-1)
        
    def erode(self, grid, dt, node_drainage_areas='drainage_area', 
                node_elevs='topographic_elevation', 
                W_if_used=None, Q_if_used=None,
                Dchar_if_used=None, io=None):
        
        """
        This method calculates the fluvial sediment transport capacity at all
        nodes, then erodes or deposits sediment as required.
        
        *grid* & *dt* are the grid object and timestep (float) respectively.
        
        *Node_drainage_areas* tells the component where to look for the drainage
        area values. Change to another string to override which grid field the
        component looks at, or pass a nnodes-long array of drainage areas values
        directly instead.
        
        If you already have slopes defined at nodes on the grid, pass them to
        the component with *slopes_at_nodes*. The same syntax is expected: 
        string gives a name in the grid fields, an array gives values direct.
        
        Alternatively, set *link_slopes* (and *link_node_mapping*) if this data
        is only available at links. 'planet_surface__derivative_of_elevation'
        is the default field name for link slopes. Override this name by
        setting the variable as the appropriate string, or override use of 
        grid fields altogether by passing an array. *link_node_mapping* controls
        how the component maps these link values onto the arrays. We assume 
        there is always a 1:1 mapping (pass the values already projected onto 
        the nodes using slopes_at_nodes if not). Other components, e.g.,
        flow_routing.route_flow_dn, may provide the necessary outputs to make
        the mapping easier: e.g., just pass 'links_to_flow_reciever' from that
        module (the default name). If the component cannot find an existing
        mapping through this parameter, it will derive one on the fly, at
        considerable cost of speed (see on-screen reports).
        
        *slopes_from_elevs* allows the module to create gradients internally
        from elevations rather than have them provided. Set to True to force 
        the component to look for the data in grid.at_node['topographic_elevation'];
        set to 'name_of_field' to override this name, or pass an nnode-array
        to use those values as elevations instead. Using this option is 
        considerably slower than any of the alternatives, as it also has to 
        calculate the link_node_mapping from stratch each time.
        
        In both these cases, at present the mapping is to use the maximum
        slope of --any-- link attached to the node as the representative node
        slope. This is primarily for speed, but may be a good idea to modify
        later.
        
        *W_if_used* and *Q_if_used* must be provided if you set use_W and use_Q
        respectively in the component initialization. They can be either field
        names or nnodes arrays as in the other cases.
        
        *Dchar_if_used* must be set as a grid field string or nnoodes-long array
        if 'Dchar' as a float was not provided in the input file. (If it was,
        this will be overridden).
        
        RETURNS XXXXXX
        """
        dx = grid.node_spacing_horizontal
        dy = grid.node_spacing_vertical
        dx2 = self.dx2
        dy2 = self.dy2
        nrows = grid.number_of_node_rows
        ncols = grid.number_of_node_columns
        
        try:
            self.Dchar=self.Dchar_in
        except AttributeError:
            try:
                self.Dchar=grid.at_node[Dchar_if_used]
            except FieldError:
                assert type(Dchar_if_used)==np.ndarray
                self.Dchar=Dchar_if_used
            
        if type(node_elevs)==str:
            node_z = grid.at_node[node_elevs]
        else:
            node_z = node_elevs
        node_z_asgrid = node_z.view().reshape((nrows,ncols))
        
        if type(node_drainage_areas)==str:
            node_A = grid.at_node[node_drainage_areas]
        else:
            node_A = node_drainage_areas
            
        ###########this still "feels" the influence of closed nodes. Need to
        ###########get rid of them.

        all_nodes_diffusivity = self.diffusivity_prefactor*node_A**self.diffusivity_power_on_A
        #########ALT
        neighbor_nodes = grid.get_neighbor_list(bad_index=-1)
        #the -1 lets us get *some* value for all nodes, which we then mask:
        neighbor_diffusivities = np.ma.array(all_nodes_diffusivity[neighbor_nodes], mask=self.bad_neighbor_mask)
        #pylab.figure(1)
        #pylab.imshow(neighbor_diffusivities[:,3].reshape((nrows,ncols)))
        #pylab.colorbar()
        #order is E,N,W,S
        mean_diffusivities_byspacing = neighbor_diffusivities+all_nodes_diffusivity.reshape((grid.number_of_nodes,1)) #not yet _byspacing...
        mean_diffusivities_byspacing[:,[0,2]] /= (2.*dx)
        mean_diffusivities_byspacing[:,[1,3]] /= (2.*dy) #now complete
        #for in the loop
        rate_of_z_change_store = np.ma.empty((mean_diffusivities_byspacing.shape[0],2),dtype=float)
        #set up the Von Neumann stability criterion:
        np.sum(mean_diffusivities_byspacing[:,[0,2]]/dx, axis=1, out=rate_of_z_change_store[:,0])
        np.sum(mean_diffusivities_byspacing[:,[1,3]]/dy, axis=1, out=rate_of_z_change_store[:,1])
        max_sum_of_Ds = np.amax(np.sum(rate_of_z_change_store, axis=1))
        
        #adjust the tstep for Von Neumann stability here:
        delta_t_internal = 1./max_sum_of_Ds
        num_reps_internal = int(dt//delta_t_internal)
        #print 'num reps: ', num_reps_internal
        dt_excess = dt%delta_t_internal
        
        for reps in xrange(num_reps_internal+1):
            #check if last loop
            if reps==num_reps_internal:
                delta_t_internal = dt_excess
        
            node_gradients = grid.calculate_gradient_along_node_links(node_z)
            #pylab.figure(2)
            #pylab.imshow(node_gradients[:,3].reshape((nrows,ncols)))
            #pylab.colorbar()
            #pylab.show()
            #this method returns (nnodes,4), & masked values where links are inactive
            #ordering is E,N,W,S, same as nieghbors and hence Ds
            D_slope_product = node_gradients*mean_diffusivities_byspacing #the masks shoud be propagating forward still
            np.sum(D_slope_product[:,:2], axis=1, out=rate_of_z_change_store[:,0])
            np.sum(-D_slope_product[:,2:], axis=1, out=rate_of_z_change_store[:,1])
            rate_of_z_change = np.sum(rate_of_z_change_store, axis=1) #this use of sum is necessary to preserve the right masking
            node_z[grid.core_nodes] += delta_t_internal*rate_of_z_change[grid.core_nodes]
        
        ##########
        #all_nodes_diffusivity_asgrid = all_nodes_diffusivity.view().reshape((nrows,ncols)) #view prevents a copy from being made accidentally
        #interior_nodes_diffusivity = all_nodes_diffusivity_asgrid[1:-1,1:-1]
        ##the "subsets" are only defined on the interior nodes (non-perimeter)
        #D_Ebydx2 = (all_nodes_diffusivity_asgrid[2:,1:-1]+interior_nodes_diffusivity)/2./dx2
        #D_Wbydx2 = (all_nodes_diffusivity_asgrid[:-2,1:-1]+interior_nodes_diffusivity)/2./dx2
        #D_Nbydy2 = (all_nodes_diffusivity_asgrid[1:-1,:-2]+interior_nodes_diffusivity)/2./dy2 #remember, the grid is upside down by this reshape
        #D_Sbydy2 = (all_nodes_diffusivity_asgrid[1:-1,2:]+interior_nodes_diffusivity)/2./dy2
        #sum_of_Ds = D_Ebydx2 + D_Wbydx2 + D_Nbydy2 + D_Sbydy2
        #        
        ##adjust the tstep for Von Neumann stability here:
        #delta_t_internal = 1./np.amax(sum_of_Ds)
        #num_reps_internal = int(dt//delta_t_internal)
        #print 'num reps: ', num_reps_internal
        #dt_excess = dt%delta_t_internal 
        ##we won't be adjusting the flow paths during step subdivision, so this fortunately stays linear, and we only do it once...
        #one_less_delta_t_times_sumofDs = 1. - delta_t_internal*sum_of_Ds
        #
        #for reps in xrange(num_reps_internal+1):
        #    #check if last loop
        #    if reps==num_reps_internal:
        #        delta_t_internal = dt_excess
        #                
        #    z_ij = node_z_asgrid[1:-1,1:-1]
        #    z_iplus1j = node_z_asgrid[2:,1:-1]
        #    z_iminus1j = node_z_asgrid[:-2,1:-1]
        #    z_ijplus1 = node_z_asgrid[1:-1,:-2]
        #    z_ijminus1 = node_z_asgrid[1:-1,2:]
        #
        #    #new_z_internal = one_less_delta_t_times_sumofDs * z_ij + delta_t_internal*(D_Ebydx2*z_iplus1j
        #    #                                                                          +D_Wbydx2*z_iminus1j
        #    #                                                                          +D_Nbydy2*z_ijplus1
        #    #                                                                          +D_Sbydy2*z_ijminus1)
        #    new_z_internal = z_ij + delta_t_internal*(D_Ebydx2*(z_iplus1j-z_ij)-D_Wbydx2*(z_ij-z_iminus1j)+D_Nbydy2*(z_ijplus1-z_ij)-D_Sbydy2*(z_ij-z_ijminus1))
        #    print 'mean_diff', np.mean(np.fabs(node_z_asgrid[1:-1,1:-1]-new_z_internal))
        #    print 'max_diff', np.max(np.fabs(node_z_asgrid[1:-1,1:-1]-new_z_internal))
        #    #we don't add uplift as we do it manually outside the component
        #    node_z_asgrid[1:-1,1:-1] = new_z_internal
        #    #repin the BCs, in case there are any interior boundaries
        #    #grid.at_node[node_elevs][sgrid.interior_nodes((nrows,ncols))] = node_z_asgrid.ravel()
        #    #grid.update_boundary_nodes()
        #    #...remember, because everything has hopefully been passed by reference, the 
        #    #fields should already have updated to reflect changes to z here
        #
        ##to see if this is actually necessary
        ##grid.at_node[node_elevs][sgrid.interior_nodes((nrows,ncols))] = node_z_asgrid.ravel()
        self.grid=grid

        active_nodes = grid.get_active_cell_node_ids()
        if io:
            try:
                io[active_nodes] += node_z_asgrid.ravel()[active_nodes]
            except TypeError:
                if type(io)==str:
                    elev_name = io
            else:
                return grid, io
            
        else:
            elev_name = 'topographic_elevation'
        
        return grid, grid.at_node[elev_name], all_nodes_diffusivity
        



#! /usr/bin/env python
"""
Unit tests for landlab.collections
"""
from nose.tools import assert_equal, assert_true, assert_raises
try:
    from nose.tools import (assert_dict_equal, assert_tuple_equal,
                            assert_list_equal)
except ImportError:
    from landlab.testing.tools import (assert_dict_equal, assert_tuple_equal,
                                       assert_list_equal)

from landlab import Palette, Arena, NoProvidersError
from landlab.components.sample import Sample1, Sample2

def test_empty_palette():
    """
    Create a palette without components.
    """
    palette = Palette()

    assert_equal(len(palette), 0)
    assert_equal(palette.list(), [])
    assert_equal(palette.keys(), [])
    assert_equal(palette.uses(), [])
    assert_equal(palette.provides(), [])

    providers = palette.find_provider('air__temperature')
    assert_equal(providers, [])

    users = palette.find_user('air__temperature')
    assert_equal(users, [])

    connections = palette.find_connections()
    assert_dict_equal(connections, {})


def test_1_component_create():
    palette = Palette(sample=Sample1)

    assert_equal(len(palette), 1)


def test_1_component_dict_interface():
    palette = Palette(sample=Sample1)

    assert_dict_equal(dict(sample=Sample1), palette)
    assert_equal(len(palette), 1)
    assert_equal(palette.keys(), ['sample'])
    assert_equal(palette.values(), [Sample1])

    items = palette.items()
    assert_tuple_equal(('sample', Sample1), items[0])


def test_1_component_list():
    palette = Palette(sample=Sample1)

    assert_equal(['sample'], palette.list())


def test_1_component_uses():
    palette = Palette(sample=Sample1)

    uses = palette.uses()
    uses.sort()
    assert_equal(['air__temperature', 'surface__elevation'], uses)


def test_1_component_provides():
    palette = Palette(sample=Sample1)

    provides = palette.provides()
    provides.sort()
    assert_equal(['deposition__rate'], provides)


def test_1_component_find_providers():
    palette = Palette(sample=Sample1)

    providers = palette.find_provider('air__temperature')
    assert_equal(providers, [])

    providers = palette.find_provider('deposition__rate')
    assert_equal(providers, ['sample'])


def test_1_component_find_users():
    palette = Palette(sample=Sample1)

    users = palette.find_user('air__temperature')
    assert_equal(users, ['sample'])


def test_1_component_find_connections():
    palette = Palette(sample=Sample1)

    assert_raises(NoProvidersError, palette.find_connections)


def test_2_components_create():
    palette = Palette(one=Sample1, two=Sample2)

    assert_equal(len(palette), 2)


def test_2_components_dict_interface():
    palette = Palette(one=Sample1, two=Sample2)

    assert_dict_equal(dict(one=Sample1, two=Sample2), palette)
    assert_equal(len(palette), 2)

    keys = palette.keys()
    keys.sort()
    assert_list_equal(['one', 'two'], keys)

    values = palette.values()
    assert_equal(2, len(values))
    assert_true(Sample1 in values and Sample2 in values)

    items = palette.items()
    items.sort()
    assert_equal(2, len(items))
    assert_tuple_equal(('one', Sample1), items[0])
    assert_tuple_equal(('two', Sample2), items[1])


def test_2_components_list():
    palette = Palette(one=Sample1, two=Sample2)

    list = palette.list()
    list.sort()
    assert_list_equal(['one', 'two'], list)


def test_2_components_uses():
    palette = Palette(one=Sample1, two=Sample2)

    uses = palette.uses()
    uses.sort()
    assert_list_equal(['air__temperature',
                       'deposition__rate',
                       'surface__elevation'], uses)


def test_2_components_provides():
    palette = Palette(one=Sample1, two=Sample2)

    provides = palette.provides()
    provides.sort()
    assert_list_equal(['air__temperature',
                       'deposition__rate',
                       'surface__elevation'], provides)


def test_2_components_find_providers():
    palette = Palette(one=Sample1, two=Sample2)

    providers = palette.find_provider('air__temperature')
    assert_list_equal(['two'], providers)

    providers = palette.find_provider('deposition__rate')
    assert_equal(['one'], providers)


def test_2_components_find_users():
    palette = Palette(one=Sample1, two=Sample2)

    users = palette.find_user('air__temperature')
    assert_list_equal(['one'], users)


def test_2_components_find_connections():
    palette = Palette(one=Sample1, two=Sample2)

    connections = {
        'one': {'deposition__rate': ['two']},
        'two': {'air__temperature': ['one'],
                'surface__elevation': ['one']},
    }
    assert_dict_equal(connections, palette.find_connections())


def test_arena_instantiate():
    arena = Arena()
    assert_dict_equal(dict(), arena)

    arena.instantiate(Sample1, 'one')
    assert_equal(1, len(arena))
    assert_true('one' in arena)
    assert_true(isinstance(arena['one'], Sample1))

    arena.instantiate(Sample2, 'two')
    assert_equal(2, len(arena))
    assert_true('one' in arena and 'two' in arena)
    assert_true(isinstance(arena['one'], Sample1))
    assert_true(isinstance(arena['two'], Sample2))


def test_arena_connect():
    arena = Arena()
    arena.instantiate(Sample1, 'one')
    arena.instantiate(Sample2, 'two')

    arena.connect('one', 'two', 'air__temperature')
    arena.connect('one', 'two', 'surface__elevation')
    arena.connect('two', 'one', 'deposition__rate')

    dz = arena['one'].get_value('deposition__rate')
    t = arena['two'].get_value('air__temperature')
    z = arena['two'].get_value('surface__elevation')


def test_arena_walk():
    arena = Arena()
    arena.instantiate(Sample1, 'one')
    arena.instantiate(Sample2, 'two')

    arena.connect('one', 'two', 'air__temperature')
    arena.connect('one', 'two', 'surface__elevation')
    arena.connect('two', 'one', 'deposition__rate')

    tree = arena.walk('one')
    assert_list_equal(['one', 'two'], tree)



#! /usr/env/python
"""

Component that models 2D diffusion using an explicit finite-volume method.

Created July 2013 GT
Last updated August 2013 GT

"""
from __future__ import print_function

from landlab import ModelParameterDictionary
from landlab import create_and_initialize_grid
from ...core.model_parameter_dictionary import MissingKeyError

_ALPHA = 0.1   # time-step stability factor

#_VERSION = 'make_all_data'
#_VERSION = 'explicit'
_VERSION = 'pass_grid'


class DiffusionComponent(object):
    """
    This component implements linear diffusion of a field in the supplied
    ModelGrid.
    
    This components requires the following parameters be set in the input file,
    *input_stream*, set in the component initialization:
        
        'DIFMOD_UPLIFT_RATE' or 'uplift_rate', both equivalent to the uplift rate
        'DIFMOD_KD', the diffusivity to use
    
    Optional inputs are:
        
        'dt', the model timestep (assumed constant)
        'values_to_diffuse', a string giving the name of the grid field
        containing the data to diffuse.
    
    If 'dt' is not supplied, you must call the method :func:`set_timestep` as
    part of your run loop. This allows you to set a dynamic timestep for this
    class.
    If 'values_to_diffuse' is not provided, defaults to 
    'topographic_elevation'.
    
    No particular units are necessary where they are not specified, as long as
    all units are internally consistent.
    
    The component takes *grid*, the ModelGrid object, and (optionally)
    *current_time* and *input_stream*. If *current_time* is not set, it defaults
    to 0.0. If *input_stream* is not set in instantiation of the class,
    :func:`initialize` with *input_stream* as in input must be called instead.
    *Input_stream* is the filename of (& optionally, path to) the parameter 
    file.
    
    The primary method of this class is :func:`diffuse`.
    """
    
    def __init__(self, grid, input_stream=None, current_time=0.):
        self.grid = grid
        self.current_time = current_time
        if input_stream:
            self.initialize(input_stream)
        else:
            print('Ensure you call the initialize(input_stream) method before '
                  'running the model!')
        
    def initialize(self, input_stream):
        
        # Create a ModelParameterDictionary for the inputs
        if type(input_stream)==ModelParameterDictionary:
            inputs = input_stream
        else:
            inputs = ModelParameterDictionary(input_stream)
        
        # Read input/configuration parameters
        self.kd = inputs.read_float('DIFMOD_KD')
        try:
            self.uplift_rate = inputs.read_float('DIFMOD_UPLIFT_RATE')
        except MissingKeyError:
            self.uplift_rate = inputs.read_float('uplift_rate')
        try:
            self.values_to_diffuse = inputs.read_str('values_to_diffuse')
        except MissingKeyError:
            self.values_to_diffuse = 'topographic_elevation'
        try:
            self.timestep_in = inputs.read_float('dt')  
        except MissingKeyError:
            print('No fixed timestep supplied, it must be set dynamically '
                  'somewhere else. Be sure to call '
                  'input_timestep(timestep_in) as part of your run loop.')

        
        # Create grid if one doesn't already exist
        if self.grid is None:
            self.grid = create_and_initialize_grid(input_stream)
            
        # Set internal time step
        # ..todo:
        #   implement mechanism to compute time-steps dynamically if grid is 
        #   adaptive/changing
        dx = self.grid.min_active_link_length()  # smallest active link length
        self.dt = _ALPHA*dx*dx/self.kd  # CFL condition
        try:
            self.tstep_ratio = self.timestep_in/self.dt
        except AttributeError:
            pass
        
        # Get a list of interior cells
        self.interior_cells = self.grid.get_core_cell_node_ids()
        
        # Here we're experimenting with different approaches: with 
        # 'make_all_data', we create and manage all the data we need and embed
        # it all in the grid. With 'explicit', we require the caller/user to 
        # provide data.
        if _VERSION=='make_all_data':
            #print('creating internal data')
            self.z = self.grid.add_zeros('node', 'landscape_surface__elevation')
            self.g = self.grid.add_zeros('active_link', 'landscape_surface__gradient')  # surface gradients
            self.qs = self.grid.add_zeros('active_link','unit_sediment_flux')  # unit sediment flux
            self.dqds = self.grid.add_zeros('node', 'sediment_flux_divergence')  # sed flux derivative
        elif _VERSION=='explicit':
            pass
        else:
            # Create data arrays for variables that won't (?) be shared with other
            # components
            self.g = self.grid.create_active_link_array_zeros()  # surface gradients
            self.qs = self.grid.create_active_link_array_zeros()  # unit sediment flux
            self.dqds = self.grid.create_node_array_zeros()  # sed flux derivative
    
    
    def input_timestep(self, timestep_in):
        """
        Allows the user to set a dynamic (evolving) timestep manually as part of
        a run loop.
        """
        self.timestep_in = timestep_in
        self.tstep_ratio = timestep_in/self.dt
        
        
    def run_one_step_explicit(self, mg, z, g, qs, dqsds, dzdt, delt):
        
        # Take the smaller of delt or built-in time-step size self.dt
        dt = min(self.dt, delt)
        
        # Calculate the gradients and sediment fluxes
        g = mg.calculate_gradients_at_active_links(z)
        qs = -self.kd*g
        
        # Calculate the net deposition/erosion rate at each node
        dqsds = mg.calculate_flux_divergence_at_nodes(qs)
        
        # Calculate the total rate of elevation change
        dzdt = self.uplift_rate - dqsds
        
        # Update the elevations
        z[self.interior_cells] = z[self.interior_cells] \
                                 + dzdt[self.interior_cells] * dt
                                 
        # Update current time and return it
        self.current_time += dt
        
        return z, g, qs, dqsds, dzdt
        
        
    def run_one_step_internal(self, delt):
        
        # Take the smaller of delt or built-in time-step size self.dt
        dt = min(self.dt, delt)
        
        # Calculate the gradients and sediment fluxes
        self.g = self.grid.calculate_gradients_at_active_links(self.z)
        self.qs = -self.kd*self.g
        
        # Calculate the net deposition/erosion rate at each node
        self.dqsds = self.grid.calculate_flux_divergence_at_nodes(self.qs)
        
        # Calculate the total rate of elevation change
        dzdt = self.uplift_rate - self.dqsds
        
        # Update the elevations
        self.z[self.interior_cells] += dzdt[self.interior_cells] * dt
                                 
        # Update current time and return it
        self.current_time += dt
        return self.current_time
    
    def diffuse(self, grid, internal_uplift=True, num_uplift_implicit_comps = 1):
        """
        This is the primary method of the class. Call it to perform an iteration
        of the model. Takes *grid*, the model grid.
        
        *grid* must contain the field to diffuse, which defaults to
        'topographic_elevation'. This can be overridden with the 
        values_to_diffuse property in the input file.
        
        See the class docstring for a list of the other properties necessary
        in the input file for this component to run.
        
        By default, this component requires it to incorporate uplift into its 
        execution. If you only have one module that requires this, do not add
        uplift manually in your loop; this method will include uplift 
        automatically. If more than one of your components has this requirement,
        set *num_uplift_implicit_comps* to the total number of components that
        do.
        
        You can suppress this behaviour by setting *internal_uplift* to False.

        """
        # Take the smaller of delt or built-in time-step size self.dt
        repeats = int(self.tstep_ratio//1.)
        extra_time = self.tstep_ratio-repeats
        z = grid.at_node[self.values_to_diffuse]
        
        core_nodes = grid.get_core_cell_node_ids()
        
        for i in xrange(repeats+1):
            # Calculate the gradients and sediment fluxes
            self.qs = -self.kd*grid.calculate_gradients_at_active_links(z)
        
            # Calculate the net deposition/erosion rate at each node
            self.dqsds = grid.calculate_flux_divergence_at_nodes(self.qs)    
            
            # Calculate the total rate of elevation change
            #dzdt = self.uplift_rate - self.dqsds
            dzdt = - self.dqsds
            
            # Update the elevations
            if i == (repeats):
                timestep = extra_time
            else:
                timestep = self.dt
            if internal_uplift:
                add_uplift = self.uplift_rate/num_uplift_implicit_comps
            else:
                add_uplift = 0.
            grid.at_node[self.values_to_diffuse][core_nodes] += add_uplift + dzdt[core_nodes] * timestep
            
            #check the BCs, update if fixed gradient
            if grid.fixed_gradient_boundary_nodes:
                grid.at_node[self.values_to_diffuse][grid.fixed_gradient_node_properties['boundary_node_IDs']] = grid.at_node[self.values_to_diffuse][grid.fixed_gradient_node_properties['anchor_node_IDs']] + grid.fixed_gradient_node_properties['values_to_add']

        #return the grid
        return grid

        
    def run_until_explicit(self, mg, t, z, g, qs, dqsds, dzdt):
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            z, g, qs, dqsds, dzdt = self.run_one_step_explicit(mg, z, g, qs, dqsds, dzdt, remaining_time)
            
        return z, g, qs, dqsds, dzdt
        
        
    def run_until_internal(self, t):
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            self.run_one_step_internal(remaining_time)
                    
        
    def run_until(self, t):  # this is just a temporary duplicate
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            self.run_one_step_internal(remaining_time)
                    
        
    def get_time_step(self):
        """
        Returns time-step size.
        """
        return self.dt
    
    @property
    def time_step(self):
        """
        Returns time-step size (as a property).
        """
        return self.dt




#################################################################
##
##  'Field' concept is implemented for Radiation component.
##
##  Sai Nudurupati and Erkan Istanbulluoglu - 14May2014
#################################################################

from landlab import Component

import numpy as np

_VALID_METHODS = set(['Grid'])

def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)

class Radiation( Component ):
    """
    Landlab component that computes 1D and 2D total incident shortwave
    radiation. This code also computes relative incidence shortwave radiation
    compared to a flat surface.

    Radiation(grid, **kwds)
        Adds two 'cellular' fields on grid 'TotalShortWaveRadiation' and
        'RadiationFactor'

    Parameters:
        grid : RasterModelGrid (might work on other grids but not tested yet)

      Optional (**kwds):
        method : Currently, only default is available
        CLOUDINESS: set cloudiness value. default value is 0.0
        LATITUDE: set Latitude. default value is 34.0
        ALBEDO: set albedo. default value is 0.2
        SOLARCONSTANT: default value is 1353.0
        CLRSKYTURBIDITY: set clear sky turbidity. default value is 2.
        OPTAIRMASS: set optical air mass. default value is 0.0

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.radiation.radiation_field import Radiation
    >>> import numpy as np
    >>> grid = RasterModelGrid( 5, 4, 0.2 )
    >>> grid['node']['Elevation'] = np.random.rand( grid.number_of_nodes ) * 1000
    >>> rad = Radiation( grid )
    >>> rad.name
    'Radiation'
    >>> current_time = 0.5
    >>> rad.update( current_time )

    >>> x = grid['cell']['TotalShortWaveRadiation']
    >>> isinstance(x, np.ndarray)
    True
    >>> x.shape
    (6,)

    >>> x = grid['cell']['RadiationFactor']
    >>> isinstance(x, np.ndarray)
    True
    >>> x.shape
    (6,)
    """

    _name = 'Radiation'

    _input_var_names = set([
        'Elevation',
    ])

    _output_var_names = set([
        'TotalShortWaveRadiation',
        'RadiationFactor',
        'NetShortWaveRadiation',
    ])

    _var_units = {
        'Elevation' : 'm',
        'TotalShortWaveRadiation' : 'W/m^2',
        'RadiationFactor' : 'None',
        'NetShortWaveRadiation' : 'W/m^2',
    }

    def __init__( self, grid, **kwds ):
        self._method = kwds.pop('method', 'Grid')
        self._N = kwds.pop('CLOUDINESS', 0.2)
        self._latitude = kwds.pop('LATITUDE', 34.0)
        self._A = kwds.pop('ALBEDO', 0.2)
        self._Io = kwds.pop('SOLARCONSTANT', 1366.67)
        self._n = kwds.pop('CLRSKYTURBIDITY', 2.0)
        self._m = kwds.pop('OPTAIRMASS', 0.0)

        assert_method_is_valid(self._method)

        super(Radiation, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if not name in self.grid.at_node:
                self.grid.add_zeros('node', name, units=self._var_units[name])

        for name in self._output_var_names:
            if not name in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        if not 'Slope' in self.grid.at_cell:
            self.grid.add_zeros('cell', 'Slope', units='radians' )

        if not 'Aspect' in self.grid.at_cell:
            self.grid.add_zeros('cell', 'Aspect', units='radians' )

        self._nodal_values = self.grid['node']
        self._cell_values = self.grid['cell']
        self._slope,self._aspect = \
            grid.calculate_slope_aspect_at_nodes_Burrough(vals = 'Elevation')
        self._cell_values['Slope'] = self._slope
        self._cell_values['Aspect'] = self._aspect

    def update( self, current_time, **kwds ):

        self._t = kwds.pop('Hour', 12.)
        #self._elev = self._nodal_values['Elevation']
        self._radf = self._cell_values['RadiationFactor']
        self._Rs = self._cell_values['TotalShortWaveRadiation']
        self._Rnet = self._cell_values['NetShortWaveRadiation']

        self._julian = np.floor( ( current_time - np.floor( current_time ) )  \
                                  * 365.25 )                          # Julian day

        self._phi = np.radians(self._latitude)        # Latitude in Radians

        self._delta = 23.45 * np.radians(np.cos(2*np.pi/365                 \
                        * (172 - self._julian)))             # Declination angle


        self._tau = (self._t + 12.0) * np.pi/12.0                # Hour angle

        self._alpha = np.arcsin(np.sin(self._delta) * np.sin(self._phi)        \
                        + np.cos(self._delta) * np.cos(self._phi)              \
                        * np.cos(self._tau))                  # Solar Altitude

        if self._alpha <= 0.25 * np.pi/180.0:         # If altitude is -ve,
            self._alpha = 0.25 * np.pi/180.0       # sun is beyond the horizon

        self._Rgl = (self._Io*np.exp((-1) * self._n * (0.128 - 0.054 *        \
                    np.log10(1/np.sin(self._alpha)))*(1/np.sin(self._alpha))))
                    # Counting for Albedo, Cloudiness and Atmospheric turbidity

        self._phisun = np.arctan(-np.sin(self._tau)/(np.tan(self._delta)       \
                    *np.cos(self._phi) - np.sin(self._phi)                     \
                        *np.cos(self._tau)))
                                                                # Sun's Azhimuth

        if ( self._phisun >= 0 and -np.sin(self._tau) <= 0 ):
            self._phisun = self._phisun + np.pi

        elif ( self._phisun <= 0 and -np.sin(self._tau) >= 0 ):
            self._phisun = self._phisun + np.pi

        self._flat = np.cos(np.arctan(0)) * np.sin(self._alpha) +              \
                      np.sin(np.arctan(0)) * np.cos(self._alpha)               \
                       * np.cos(self._phisun - 0)
                                                       # flat surface reference

        self._Rsflat = self._Rgl * self._flat
                            # flat surface total incoming shortwave radiation

        self._Rnetflat = (1 - self._A) * (1 - 0.65 * (self._N**2)) * \
                                self._Rsflat
                            # flat surface Net incoming shortwave radiation

        self._sloped = (np.cos(self._slope) *                        \
                    np.sin(self._alpha) + np.sin(self._slope) *    \
                    np.cos(self._alpha) *                          \
                    np.cos(self._phisun - self._aspect))

        self._radf = self._sloped/self._flat

        self._radf[self._radf<=0.] = 0.
        self._radf[self._radf>6.] = 6.

        self._Rs = self._Rsflat * self._radf
                    # Sloped surface Toatl Incoming Shortwave Radn
        self._Rnet = self._Rnetflat * self._radf

        self._cell_values['RadiationFactor'] = self._radf
        self._cell_values['TotalShortWaveRadiation'] = self._Rs
        self._cell_values['NetShortWaveRadiation'] = self._Rnet



#!/usr/bin/env python

import numpy as np

from landlab import Component
from .funcs import get_flexure_parameter


_VALID_METHODS = set(['airy', 'flexure'])


def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class FlexureComponent(Component):
    """
    Landlab component that implements a 1 and 2D lithospheric flexure
    model.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flexure import FlexureComponent
    >>> grid = RasterModelGrid(5, 4, 1.e4)
    >>> flex = FlexureComponent(grid)
    >>> flex.name
    'Flexure'
    >>> sorted(flex.input_var_names)
    ['lithosphere__elevation', 'lithosphere__overlying_pressure', 'planet_surface_sediment__deposition_increment']
    >>> sorted(flex.output_var_names)
    ['lithosphere__elevation', 'lithosphere__elevation_increment']
    >>> for var in sorted(flex.units): flex.units[var]
    'm'
    'm'
    'Pa'
    'm'

    >>> flex.grid.number_of_node_rows
    5
    >>> flex.grid.number_of_node_columns
    4
    >>> flex.grid is grid
    True

    >>> np.all(grid.at_node['lithosphere__elevation'] == 0.)
    True

    >>> np.all(grid.at_node['lithosphere__elevation'] == 0.)
    True
    >>> np.all(grid.at_node['lithosphere__overlying_pressure'] == 0.)
    True
    >>> flex.update()
    >>> np.all(grid.at_node['lithosphere__elevation_increment'] == 0.)
    True

    >>> load = grid.at_node['lithosphere__overlying_pressure']
    >>> load[4] = 1e9
    >>> dz = grid.at_node['lithosphere__elevation_increment']
    >>> np.all(dz == 0.)
    True

    >>> flex.update()
    >>> np.all(grid.at_node['lithosphere__elevation_increment'] == 0.)
    False
    """
    
    ###############THIS IS THE STANDARD DECLARATIONS EXAMPLE.
    #YOU NEED TO DECLARE ALL THE FOLLOWING LIKE THIS HERE IN YOUR COMPONENT!!
    _name = 'Flexure'
    #...give the component a name

    _input_var_names = set([
        'lithosphere__overlying_pressure',
        'lithosphere__elevation',
        'planet_surface_sediment__deposition_increment',
    ])
    #...the component requires these values to do its calculation
    
    _output_var_names = set([
        'lithosphere__elevation_increment',
        'lithosphere__elevation',
    ])
    #...the component modifies these values

    # the next three dictionaries must each have the union of the two var_name
    # sets above as their keys
    _var_units = {
        'lithosphere__overlying_pressure': 'Pa',
        'lithosphere__elevation': 'm',
        'lithosphere__elevation_increment': 'm',
        'planet_surface_sediment__deposition_increment': 'm',
    }
    #...the units for each field. In the future, there may be unit casting
    # required, as there's nothing stopping you passing in the wrong units!!!
    
    _var_mapping = {
        'lithosphere__overlying_pressure': 'node',
        'lithosphere__elevation': 'node',
        'lithosphere__elevation_increment': 'node',
        'planet_surface_sediment__deposition_increment': 'node',
    }
    #...the grid centering of each name. Note you CANNOT and SHOULD NOT TRY
    # TO define the same field name but with different centering within one
    # component; you'll need to distinguish the names. 
    
    _var_defs = {
        'lithosphere__overlying_pressure': 'The pressure at the base of the lithosphere',
        'lithosphere__elevation': 'The elevation of the top of the lithosphere, i.e., the land surface',
        'lithosphere__elevation_increment': 'The change in elevation of the top of the lithosphere (the land surface) in one timestep',
        'planet_surface_sediment__deposition_increment': 'The amount of sediment deposited at the land surface in one timestep',
    }
    #...give the names a short description. [We still need to work out how
    # we deal with same name, but different description between components]
    
    ###lastly, ---> 1. did you remember to import and inherit from Component?
    #do the import up top ->
    #from landlab import Component
    #& in the class declaration ->
    #   class FlexureComponent(Component): 
    #        ...
    ### ---> 2. Make sure that self._grid is an alias for the grid after
    # initialization, below
    ################################

    def __init__(self, grid, **kwds):
        self._eet = kwds.pop('eet', 65000.)
        self._youngs = kwds.pop('youngs', 7e10)
        self._method = kwds.pop('method', 'airy')
        self._grid = grid

        assert_method_is_valid(self._method)

        super(FlexureComponent, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros('node', name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros('node', name, units=self._var_units[name])

        self._last_load = self.grid.field_values('node', 'lithosphere__overlying_pressure').copy()

        self._nodal_values = self.grid['node']

        self._r = self._set_kei_func_grid()

    def _set_kei_func_grid(self):
        from scipy.special import kei

        alpha = get_flexure_parameter(self._eet, self._youngs, 2)
        dx, dy = np.meshgrid(
            np.arange(self._grid.number_of_node_columns) * self._grid.dx,
            np.arange(self._grid.number_of_node_rows) * self._grid.dx)

        return kei(np.sqrt(dx ** 2 + dy ** 2) / alpha)

    def update(self, n_procs=1):
        elevation = self._nodal_values['lithosphere__elevation']
        load = self._nodal_values['lithosphere__overlying_pressure']
        deflection = self._nodal_values['lithosphere__elevation_increment']
        deposition = self._nodal_values['planet_surface_sediment__deposition_increment']

        new_load = ((load - self._last_load) +
                    (deposition * 2650. * 9.81).flat)

        self._last_load = load.copy()

        deflection.fill(0.)

        if self._method == 'airy':
            deflection[:] = new_load / (3300. * 9.81)
        else:
            self.subside_loads(new_load, deflection=deflection,
                               n_procs=n_procs)
        np.subtract(elevation, deflection, out=elevation)

    def subside_loads(self, loads, deflection=None, n_procs=1):
        if deflection is None:
            deflection = np.empty(self.shape, dtype=np.float)

        from .cfuncs import subside_grid_in_parallel

        w = deflection.reshape(self._grid.shape)
        load = loads.reshape(self._grid.shape)
        alpha = get_flexure_parameter(self._eet, self._youngs, 2)

        subside_grid_in_parallel(w, load, self._r, alpha, n_procs)

        return deflection


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#! /usr/env/python

"""
fracture_grid: creates and returns a 2D grid with randomly generated fractures.
The grid contains the value 1 where fractures (one cell wide) exist, and
0 elsewhere. The idea is to use this for simulations based on weathering and
erosion of, and/or flow within, fracture networks.

The entry point is the function:
    
    make_frac_grid(frac_spacing, numrows=50, numcols=50, model_grid=None)
    
If called with a Landlab RasterModelGrid, the function returns the fracture
grid as a node array. Otherwise, it returns a numrows x numcols Numpy array.

Potential improvements:
    - Add doctests
    - Fractures could be defined by links rather than nodes (i.e., return a
        link array with a code indicating whether the link crosses a fracture
        or not)
    - Fractures could have a finite length rather than extending all the way
        across the grid
    - Use of starting position along either x or y axis makes fracture net
        somewhat asymmetric. One would need a different algorithm to make it
        fully (statistically) symmetric.
        
Created: September 2013 by Greg Tucker
Last significant modification: August 2014 GT
"""

from numpy import *
from numpy.random import *
import random
import pylab as plt


def calculate_fracture_starting_position(numrows, numcols):
    """
    Chooses a random starting position along the x or y axis (random choice).
    
    Parameters
    ----------
    numrows, numcols : int
        Number of rows and columns in the grid
        
    Returns
    -------
    x, y : int
        Fracture starting coordinates
    """
    if random.randint(0, 1)==0:
        x = 0
        y = random.randint(0, numrows-1)
    else:
        x = random.randint(0, numcols-1)        
        y = 0
    return x, y
    
    
def calculate_fracture_orientation(x, y):
    """
    Chooses a random orientation for the fracture.
    
    Parameters
    ----------
    x, y : int
        Starting coordinates (one of which should be zero)
        
    Returns
    -------
    ang : float
        Fracture angle relative to horizontal
        
    Notes
    -----
    If the fracture starts along the bottom of the grid (y=0), then the angle
    will be between 45 and 135 degrees from horizontal (counter-clockwise).
    Otherwise, it will be between -45 and 45 degrees.
    """
    ang = (pi/2)*rand()
    if y==0:
        ang += pi/4
    else:
        ang -= pi/4

    return ang

    
def calculate_fracture_step_sizes(startx, starty, ang):
    """
    Calculates the sizes of steps dx and dy to be used when "drawing" the
    fracture onto the grid.
    
    Parameters
    ----------
    startx, starty : int
        Starting grid coordinates
    ang : float
        Fracture angle relative to horizontal (radians)
        
    Returns
    -------
    dx, dy : float
        Step sizes in x and y directions. One will always be unity, and the 
    other will always be <1.
    """
    if startx==0:  # frac starts on left side
        dx = 1
        dy = tan(ang)
    else:  # frac starts on bottom side
        dy = 1
        dx = -tan(ang-pi/2)
        
    return dx, dy
    
    
def trace_fracture_through_grid(m, x0, y0, dx, dy):
    """
    Creates a "fracture" in a 2D grid, m, by setting cell values to unity along
    the trace of the fracture (i.e., "drawing" a line throuh the grid).
    
    Parameters
    ----------
    m : 2D Numpy array
        Array that represents the grid
    x0, y0 : int
        Starting grid coordinates for fracture
    dx, dy : float
        Step sizes in x and y directions
        
    Returns
    -------
    None, but changes contents of m
    """
    x = x0
    y = y0
    
    while round(x)<size(m, 1) and round(y)<size(m, 0) \
            and round(x)>=0 and round(y)>=0:
        m[round(y),round(x)] = 1
        x += dx
        y += dy


def make_frac_grid(frac_spacing, numrows=50, numcols=50, model_grid=None):
    """
    Creates and returns a grid containing a network of random fractures, which
    are represented as 1's embedded in a grid of 0's.
    
    Parameters
    ----------
    frac_spacing : int
        Average spacing of fractures (in grid cells)
    (optional) numrows, numcols : int 
        Number of rows and columns in grid (if model_grid parameter is given,
        uses values from the model grid instead)
    (optional) model_grid : Landlab RasterModelGrid object
        RasterModelGrid to use for grid size
        
    Returns
    -------
    m : Numpy array
        Array containing fracture grid, represented as 0's (matrix) and 1's
        (fractures). If model_grid parameter is given, returns a 1D array
        corresponding to a node-based array in the model grid. Otherwise,
        returns a 2D array with dimensions given by numrows, numcols.
    """
    # Make an initial grid of all zeros. If user specified a model grid,
    # use that. Otherwise, use the given dimensions.
    if model_grid is not None:
        numrows = model_grid.number_of_node_rows
        numcols = model_grid.number_of_node_columns
    m = zeros((numrows,numcols))
    
    # Add fractures to grid
    nfracs = (numrows+numcols)/frac_spacing
    for i in range(nfracs):
        
        x, y = calculate_fracture_starting_position(numrows, numcols)
        ang = calculate_fracture_orientation(x, y)
        dx, dy = calculate_fracture_step_sizes(x, y, ang)
    
        #print 'startx=',x,'starty=',y,'ang=',180*ang/pi,'dx=',dx,'dy=',dy
    
        trace_fracture_through_grid(m, x, y, dx, dy)
    
    # If we have a model_grid, flatten the frac grid so it's equivalent to
    # a node array.
    if model_grid is not None:
        #print 'FROG!'
        m.shape = (m.shape[0]*m.shape[1])

    return m
            
    
def test_fracture_grid():
    """
    Test routine that generates and displays a fracture grid.

    Parameters
    ----------
    (none)
    
    Returns
    -------
    (none)
    """
    # User-defined parameters
    N = 100
    frac_spacing = 8
    
    frac_grid = make_frac_grid(frac_spacing, N, N)
    
    plt.figure()
    plt.imshow(frac_grid)
    plt.show()
        
    # Version with model grid
    from landlab import RasterModelGrid
    rmg = RasterModelGrid(40, 50, 1.0)
    frac_grid = make_frac_grid(frac_spacing, model_grid=rmg)
    fg_raster = rmg.node_vector_to_raster(frac_grid)
    
    plt.figure()
    plt.imshow(fg_raster)
    plt.show()
            



import numpy as np

from .status import StatusGrid
from .links import link_is_active, find_active_links, LinkGrid, _split_link_ends
from .cells import CellGrid
from .nodes import NodeGrid


def _default_axis_names(n_dims):
    """Returns a tuple of the default axis names."""
    _DEFAULT_NAMES = ('z', 'y', 'x')
    return _DEFAULT_NAMES[- n_dims:]


def _default_axis_units(n_dims):
    """Returns a tuple of the default axis units."""
    return ('-', ) * n_dims


class BaseGrid(object):
    def __init__(self, nodes, axis_name=None, axis_units=None, node_status=None,
                 links=None, cells=None):
        """__init__([coord0, coord1, ...], axis_name=None, axis_units=None)

        Parameters
        ----------
        coord0, coord1, ... : sequence of array-like
            Coordinates of grid nodes
        axis_name : sequence of strings, optional
            Names of coordinate axes
        axis_units : sequence of strings, optional
            Units of coordinate axes

        Returns
        -------
        BaseGrid :
            A newly-created BaseGrid

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> ngrid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
        >>> ngrid.number_of_nodes
        4
        >>> ngrid.x_at_node
        array([ 0.,  1.,  0.,  1.])
        >>> ngrid.x_at_node[2]
        0.0
        >>> ngrid.point_at_node[2]
        array([ 1.,  0.])
        >>> ngrid.coord_at_node[:, [2, 3]]
        array([[ 1.,  1.],
               [ 0.,  1.]])

        >>> cells = ([0, 1, 2, 1, 3, 2], [3, 3], [0, 1])
        >>> ngrid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]), cells=cells)
        >>> ngrid.number_of_cells
        2
        >>> ngrid.node_at_cell
        array([0, 1])

        >>> links = [(0, 2), (1, 3), (0, 1), (1, 2), (0, 3)]
        >>> ngrid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]), links=zip(*links))
        >>> ngrid.number_of_links
        5
        >>> ngrid.links_leaving_at_node(0)
        array([0, 2, 4])
        >>> ngrid.links_entering_at_node(0)
        array([], dtype=int64)

        >>> grid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]),
        ...     node_status=[0, 0, 0, 4], links=zip(*links))
        >>> grid.status_at_node
        array([0, 0, 0, 4])
        >>> grid.active_links_entering_at_node(0)
        array([], dtype=int64)
        >>> grid.active_links_leaving_at_node(0)
        array([0, 2])
        """
        self._node_grid = NodeGrid(nodes)

        self._axis_name = tuple(axis_name or _default_axis_names(self.ndim))
        self._axis_units = tuple(axis_units or _default_axis_units(self.ndim))

        if cells is not None:
            try:
                self._cell_grid = CellGrid(*cells)
            except TypeError:
                self._cell_grid = cells

        if links is not None:
            try:
                self._link_grid = LinkGrid(links, self.number_of_nodes)
            except TypeError:
                self._link_grid = links

        if node_status is not None:
            self._status_grid = StatusGrid(node_status)

        if links is not None and node_status is not None:
            links = _split_link_ends(links)
            self._active_link_grid = BaseGrid.create_active_link_grid(
                self.status_at_node, links, self.number_of_nodes)

    @staticmethod
    def create_active_link_grid(node_status, links, number_of_nodes):
        active_link_ids = find_active_links(node_status, links)
        return LinkGrid((links[0][active_link_ids], links[1][active_link_ids]),
                        number_of_nodes, link_ids=active_link_ids)

    @property
    def ndim(self):
        return self._node_grid.ndim

    @property
    def axis_units(self):
        """Coordinate units of each axis.

        Returns
        -------
        tuple of strings :
            Coordinate units of each axis.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> ngrid = BaseGrid(([0, 1, 0], [1, 1, 0]))
        >>> ngrid.axis_units
        ('-', '-')

        >>> ngrid = BaseGrid(([0, 1, 0], [1, 1, 0]),
        ...     axis_units=['degrees_north', 'degrees_east'])
        >>> ngrid.axis_units
        ('degrees_north', 'degrees_east')
        """
        return self._axis_units

    @property
    def axis_name(self):
        """Name of each axis.

        Returns
        -------
        tuple of strings :
            Names of each axis.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> ngrid = BaseGrid(([0, 1, 0], [1, 1, 0]))
        >>> ngrid.axis_name
        ('y', 'x')

        >>> ngrid = BaseGrid(([0, 1, 0], [1, 1, 0]), axis_name=['lat', 'lon'])
        >>> ngrid.axis_name
        ('lat', 'lon')
        """
        return self._axis_name

    @property
    def number_of_links(self):
        """Number of links.
        """
        return self._link_grid.number_of_links

    @property
    def number_of_cells(self):
        """Number of cells.
        """
        return self._cell_grid.number_of_cells

    @property
    def number_of_nodes(self):
        """Number of nodes.
        """
        return self._node_grid.number_of_nodes

    @property
    def coord_at_node(self):
        return self._node_grid.coord

    @property
    def x_at_node(self):
        return self._node_grid.x

    @property
    def y_at_node(self):
        return self._node_grid.y

    @property
    def point_at_node(self):
        return self._node_grid.point

    def links_leaving_at_node(self, node):
        return self._link_grid.out_link_at_node(node)

    def links_entering_at_node(self, node):
        return self._link_grid.in_link_at_node(node)

    def active_links_leaving_at_node(self, node):
        return self._active_link_grid.out_link_at_node(node)

    def active_links_entering_at_node(self, node):
        return self._active_link_grid.in_link_at_node(node)

    @property
    def node_at_link_start(self):
        return self._link_grid.node_at_link_start

    @property
    def node_at_link_end(self):
        return self._link_grid.node_at_link_end

    @property
    def node_at_cell(self):
        return self._cell_grid.node_at_cell

    @property
    def cell_at_node(self):
        return self._cell_grid.cell_at_node

    def core_cells(self):
        return self.cell_at_node[self.core_nodes]

    @property
    def status_at_node(self):
        return self._status_grid.node_status

    @status_at_node.setter
    def status_at_node(self, status):
        self._status_grid.node_status = status
        self._active_link_grid = BaseGrid.create_active_link_grid(
            self.status_at_node, (self.node_at_link_start,
                                  self.node_at_link_end), self.number_of_nodes)

    def active_nodes(self):
        return self._status_grid.active_nodes()

    def core_nodes(self):
        return self._status_grid.core_nodes()

    def boundary_nodes(self):
        return self._status_grid.boundary_nodes()

    def closed_boundary_nodes(self):
        return self._status_grid.closed_boundary_nodes()

    def fixed_gradient_boundary_nodes(self):
        return self._status_grid.fixed_gradient_boundary_nodes()

    def fixed_value_boundary_nodes(self):
        return self._status_grid.fixed_value_boundary_nodes()

    def active_links(self):
        return self._active_link_grid.link_id

    def link_length(self, link=None):
        """Length of grid links.

        Parameters
        ----------
        link : array-like, optional
            Link IDs

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> links = [(0, 2), (1, 3), (0, 1), (2, 3), (0, 3)]
        >>> grid = BaseGrid(([0, 0, 4, 4], [0, 3, 0, 3]), links=zip(*links))
        >>> grid.link_length()
        array([ 4.,  4.,  3.,  3.,  5.])
        >>> grid.link_length(0)
        array([ 4.])

        >>> grid.link_length().min()
        3.0
        >>> grid.link_length().max()
        5.0
        """
        if link is None:
            node0, node1 = (self.node_at_link_start, self.node_at_link_end)
        else:
            node0, node1 = (self.node_at_link_start[link],
                            self.node_at_link_end[link])

        return self.node_to_node_distance(node0, node1)

    def node_to_node_distance(self, node0, node1, out=None):
        """Distance between nodes.

        Parameters
        ----------
        node0 : array-like
            Node ID of start
        node1 : array-like
            Node ID of end

        Returns
        -------
        array :
            Distances between nodes.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> grid = BaseGrid(([0, 0, 4, 4], [0, 3, 0, 3]))
        >>> grid.node_to_node_distance(0, 3)
        array([ 5.])
        >>> grid.node_to_node_distance(0, [0, 1, 2, 3])
        array([ 0.,  3.,  4.,  5.])
        """
        return point_to_point_distance(
            self._get_coord_at_node(node0), self._get_coord_at_node(node1), out=out)

        node0, node1 = np.broadcast_arrays(node0, node1)
        return np.sqrt(np.sum((self.coord_at_node[:, node1] -
                               self.coord_at_node[:, node0]) ** 2, axis=0))

    def point_to_node_distance(self, point, node=None, out=None):
        """Distance from a point to a node.

        Parameters
        ----------
        point : tuple
            Coordinates of point
        node : array-like
            Node IDs

        Returns
        -------
        array :
            Distances from point to node.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> grid = BaseGrid(([0, 0, 4, 4], [0, 3, 0, 3]))
        >>> grid.point_to_node_distance((0., 0.), [1, 2, 3])
        array([ 3.,  4.,  5.])
        >>> grid.point_to_node_distance((0., 0.))
        array([ 0.,  3.,  4.,  5.])
        >>> out = np.empty(4)
        >>> out is grid.point_to_node_distance((0., 0.), out=out)
        True
        >>> out
        array([ 0.,  3.,  4.,  5.])
        """
        return point_to_point_distance(point, self._get_coord_at_node(node), out=out)

    def point_to_node_angle(self, point, node=None, out=None):
        """Angle from a point to a node.

        Parameters
        ----------
        point : tuple
            Coordinates of point
        node : array-like
            Node IDs

        Returns
        -------
        array :
            Angles from point to node as radians.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> grid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
        >>> grid.point_to_node_angle((0., 0.), [1, 2, 3]) / np.pi
        array([ 0.  ,  0.5 ,  0.25])
        >>> grid.point_to_node_angle((0., 0.)) / np.pi
        array([ 0.  ,  0.  ,  0.5 ,  0.25])
        >>> out = np.empty(4)
        >>> out is grid.point_to_node_angle((0., 0.), out=out)
        True
        >>> out / np.pi
        array([ 0.  ,  0.  ,  0.5 ,  0.25])
        """
        return point_to_point_angle(point, self._get_coord_at_node(node), out=out)

    def point_to_node_azimuth(self, point, node=None, out=None):
        """Azimuth from a point to a node.

        Parameters
        ----------
        point : tuple
            Coordinates of point
        node : array-like
            Node IDs

        Returns
        -------
        array :
            Azimuths from point to node.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> grid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
        >>> grid.point_to_node_azimuth((0., 0.), [1, 2, 3])
        array([ 90.,   0.,  45.])
        >>> grid.point_to_node_azimuth((0., 0.))
        array([ 90.,  90.,   0.,  45.])
        >>> grid.point_to_node_azimuth((0., 0.), 1)
        array([ 90.])
        >>> out = np.empty(4)
        >>> out is grid.point_to_node_azimuth((0., 0.), out=out)
        True
        >>> out
        array([ 90.,  90.,   0.,  45.])
        """
        return point_to_point_azimuth(point, self._get_coord_at_node(node), out=out)

    def point_to_node_vector(self, point, node=None, out=None):
        """Azimuth from a point to a node.

        Parameters
        ----------
        point : tuple
            Coordinates of point
        node : array-like
            Node IDs

        Returns
        -------
        array :
            Vector from point to node.

        Examples
        --------
        >>> from landlab.grid.unstructured.base import BaseGrid
        >>> grid = BaseGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
        >>> grid.point_to_node_vector((0., 0.), [1, 2, 3])
        array([[ 0.,  1.,  1.],
               [ 1.,  0.,  1.]])
        >>> grid.point_to_node_vector((0., 0.))
        array([[ 0.,  0.,  1.,  1.],
               [ 0.,  1.,  0.,  1.]])
        >>> grid.point_to_node_vector((0., 0.), 1)
        array([[ 0.],
               [ 1.]])
        >>> out = np.empty((2, 1))
        >>> out is grid.point_to_node_vector((0., 0.), 1, out=out)
        True
        >>> out
        array([[ 0.],
               [ 1.]])
        """
        return point_to_point_vector(point, self._get_coord_at_node(node), out=out)

    def _get_coord_at_node(self, node=None):
        if node is None:
            return self.coord_at_node
        else:
            return self.coord_at_node[:, node].reshape((2, -1))


def point_to_point_distance(point0, point1, out=None):
    """Length of vector that joins two points.

    Parameters
    ----------
    (y0, x0) : tuple of array_like
    (y1, x1) : tuple of array_like
    out : array_like, optional
        An array to store the output. Must be the same shape as the output would
        have.

    Returns
    -------
    l : array_like
        Length of vector joining points; if *out* is provided, *v* will be equal
        to *out*.

    Examples
    --------
    >>> from landlab.grid.unstructured.base import point_to_point_distance
    >>> point_to_point_distance((0, 0), (3, 4))
    array([ 5.])
    >>> point_to_point_distance((0, 0), ([3, 6], [4, 8]))
    array([  5.,  10.])
    """
    point0 = np.reshape(point0, (2, -1))
    point1 = np.reshape(point1, (2, -1))
    if out is None:
        sum_of_squares = np.sum((point1 - point0) ** 2., axis=0)
        return np.sqrt(sum_of_squares)
    else:
        sum_of_squares = np.sum((point1 - point0) ** 2., axis=0, out=out)
        return np.sqrt(sum_of_squares, out=out)


def point_to_point_angle(point0, point1, out=None):
    """Angle of vector that joins two points.

    Parameters
    ----------
    (y0, x0) : tuple of array_like
    (y1, x1) : tuple of array_like
    out : array_like, optional
        An array to store the output. Must be the same shape as the output would
        have.

    Returns
    -------
    a : array_like
        Angle of vector joining points; if *out* is provided, *v* will be equal
        to *out*.
    """
    point0 = np.reshape(point0, (-1, 1))
    diff = point_to_point_vector(point0, point1)
    if out is None:
        return np.arctan2(diff[0], diff[1])
    else:
        return np.arctan2(diff[0], diff[1], out=out)


def point_to_point_azimuth(point0, point1, out=None):
    """Azimuth of vector that joins two points.

    Parameters
    ----------
    (y0, x0) : tuple of array_like
    (y1, x1) : tuple of array_like
    out : array_like, optional
        An array to store the output. Must be the same shape as the output would
        have.

    Returns
    -------
    azimuth : array_like
        Azimuth of vector joining points; if *out* is provided, *v* will be
        equal to *out*.

    Examples
    --------
    >>> from landlab.grid.unstructured.base import point_to_point_azimuth
    >>> point_to_point_azimuth((0, 0), (1, 0))
    array([ 0.])
    >>> point_to_point_azimuth([(0, 1), (0, 1)], (1, 0))
    array([  0., -90.])
    >>> point_to_point_azimuth([(0, 1, 0), (0, 1, 2)], [(1, 1, 2), (0, 0, 4)])
    array([  0., -90.,  45.])
    """
    azimuth_in_rads = point_to_point_angle(point0, point1, out=out)
    if out is None:
        return (np.pi * .5 - azimuth_in_rads) * 180. / np.pi
    else:
        np.subtract(np.pi * .5, azimuth_in_rads, out=out)
        return np.multiply(out, 180. / np.pi, out=out)


def point_to_point_vector(point0, point1, out=None):
    """Vector that joins two points.

    Parameters
    ----------
    (y0, x0) : tuple of array_like
    (y1, x1) : tuple of array_like
    out : array_like, optional
        An array to store the output. Must be the same shape as the output would
        have.

    Returns
    -------
    (dy, dx) : tuple of array_like
        Vectors between points; if *out* is provided, *v* will be equal to
        *out*.

    Examples
    --------
    >>> from landlab.grid.unstructured.base import point_to_point_vector
    >>> point_to_point_vector((0, 0), (1, 2))
    array([[1],
           [2]])
    >>> point_to_point_vector([(0, 1), (0, 1)], (1, 2))
    array([[1, 0],
           [2, 1]])
    >>> point_to_point_vector([(0, 0, 0), (0, 1, 2)], [(1, 2, 2), (2, 4, 4)])
    array([[1, 2, 2],
           [2, 3, 2]])
    """
    point0 = np.reshape(point0, (2, -1))
    point1 = np.reshape(point1, (2, -1))

    if out is None:
        return np.subtract(point1, point0)
    else:
        return np.subtract(point1, point0, out=out)



#! /usr/bin/env python

import types

from landlab.core import model_parameter_dictionary as mpd
from .raster import from_dict as raster_from_dict
from .hex import from_dict as hex_from_dict


class Error(Exception):
    pass


class BadGridTypeError(Error):
    def __init__(self, grid_type):
        self._type = str(grid_type)

    def __str__(self):
        return self._type


_GRID_READERS = {
    'raster': raster_from_dict,
    'hex': hex_from_dict,
}


def create_and_initialize_grid(input_source):
    """
    Creates, initializes, and returns a new grid object using parameters 
    specified in *input_source*. *input_source* is either a
    ModelParameterDictionary instance (or, really, just dict-like) or a
    named input file.
    
    Example:
        
    >>> from StringIO import StringIO
    >>> test_file = StringIO('''
    ... GRID_TYPE:
    ... raster
    ... NUM_ROWS:
    ... 4
    ... NUM_COLS:
    ... 5
    ... GRID_SPACING: 
    ... 2.5
    ... ''')
    >>> from landlab import create_and_initialize_grid
    >>> mg = create_and_initialize_grid(test_file)
    >>> mg.number_of_nodes
    20
        
    """    
    # Handle input source. 
    #In this code block, we do the following:
    #   - handle the case in which caller provides neither a parameter
    #     dictionary nor an input file name.
    #   - if we're given an input file name, create a parameter dictionary
    #     object that reads the specified file name
    if isinstance(input_source, dict):
        param_dict = input_source
    else:
        param_dict = mpd.ModelParameterDictionary(from_file=input_source)
        
    # Find out what type of grid the user wants
    #
    # Dev note: could handle defaults like:
    #    param_dict.get('GRID_TYPE','raster')
    # so if no GRID_TYPE is specified you get the second arg as default. If no
    # second arg, then exception
    try:
        grid_type = param_dict['GRID_TYPE']
    except KeyError:
        raise

    grid_type.strip().lower()   # make LC w/o leading/trailing spaces
    
    # Read parameters appropriate to that type, create it, and initialize it
    try:
        grid_reader = _GRID_READERS[grid_type]
    except KeyError:
        raise BadGridTypeError(grid_type)
    else:
        mg = grid_reader(param_dict)
    
    # Return the created and initialized grid
    return mg



from . import links


def number_of_faces(shape):
    """Number of faces in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of faces in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.faces import number_of_faces
    >>> number_of_faces((3, 4))
    17
    """
    return links.number_of_links(shape)
